from zfs_autobackup.CmdPipe import CmdPipe

from basetest import *
import time

from zfs_autobackup.LogConsole import  LogConsole


class TestZfsAutobackup(unittest2.TestCase):

    def setUp(self):
        prepare_zpools()
        self.longMessage=True

    def test_invalidpars(self):

        self.assertEqual(ZfsAutobackup("test test_target1 --no-progress --keep-source -1".split(" ")).run(), 255)

        with OutputIO() as buf:
            with redirect_stdout(buf):
                self.assertEqual(ZfsAutobackup("test test_target1 --no-progress --resume --verbose --no-snapshot".split(" ")).run(), 0)

            print(buf.getvalue())
            self.assertIn("The --resume", buf.getvalue())

        with OutputIO() as buf:
            with redirect_stderr(buf):
                self.assertEqual(ZfsAutobackup("test test_target_nonexisting --no-progress".split(" ")).run(), 255)

            print(buf.getvalue())
            # correct message?
            self.assertIn("Please create this dataset", buf.getvalue())


    def  test_snapshotmode(self):
        """test snapshot tool mode"""

        with mocktime("20101111000000"):
            self.assertFalse(ZfsAutobackup("test --no-progress --verbose".split(" ")).run())

        r=shelltest("zfs list -H -o name -r -t all "+TEST_POOLS)
        self.assertMultiLineEqual(r,"""
test_source1
test_source1/fs1
test_source1/fs1@test-20101111000000
test_source1/fs1/sub
test_source1/fs1/sub@test-20101111000000
test_source2
test_source2/fs2
test_source2/fs2/sub
test_source2/fs2/sub@test-20101111000000
test_source2/fs3
test_source2/fs3/sub
test_target1
""")

    def  test_defaults(self):
        self.maxDiff=2000

        with self.subTest("no datasets selected"):
            with OutputIO() as buf:
                with redirect_stderr(buf):
                    with mocktime("20101111000000"):
                        self.assertTrue(ZfsAutobackup("nonexisting test_target1 --verbose --debug --no-progress".split(" ")).run())

                print(buf.getvalue())
                #correct message?
                self.assertIn("No source filesystems selected", buf.getvalue())


        with self.subTest("defaults with full verbose and debug"):

            with mocktime("20101111000000"):
                self.assertFalse(ZfsAutobackup("test test_target1 --verbose --debug --no-progress".split(" ")).run())

            r=shelltest("zfs list -H -o name -r -t all "+TEST_POOLS)
            self.assertMultiLineEqual(r,"""
test_source1
test_source1/fs1
test_source1/fs1@test-20101111000000
test_source1/fs1/sub
test_source1/fs1/sub@test-20101111000000
test_source2
test_source2/fs2
test_source2/fs2/sub
test_source2/fs2/sub@test-20101111000000
test_source2/fs3
test_source2/fs3/sub
test_target1
test_target1/test_source1
test_target1/test_source1/fs1
test_target1/test_source1/fs1@test-20101111000000
test_target1/test_source1/fs1/sub
test_target1/test_source1/fs1/sub@test-20101111000000
test_target1/test_source2
test_target1/test_source2/fs2
test_target1/test_source2/fs2/sub
test_target1/test_source2/fs2/sub@test-20101111000000
""")

        with self.subTest("bare defaults, allow empty"):
            with mocktime("20101111000001"):
                self.assertFalse(ZfsAutobackup("test test_target1 --allow-empty --no-progress".split(" ")).run())


            r=shelltest("zfs list -H -o name -r -t all "+TEST_POOLS)
            self.assertMultiLineEqual(r,"""
test_source1
test_source1/fs1
test_source1/fs1@test-20101111000000
test_source1/fs1@test-20101111000001
test_source1/fs1/sub
test_source1/fs1/sub@test-20101111000000
test_source1/fs1/sub@test-20101111000001
test_source2
test_source2/fs2
test_source2/fs2/sub
test_source2/fs2/sub@test-20101111000000
test_source2/fs2/sub@test-20101111000001
test_source2/fs3
test_source2/fs3/sub
test_target1
test_target1/test_source1
test_target1/test_source1/fs1
test_target1/test_source1/fs1@test-20101111000000
test_target1/test_source1/fs1@test-20101111000001
test_target1/test_source1/fs1/sub
test_target1/test_source1/fs1/sub@test-20101111000000
test_target1/test_source1/fs1/sub@test-20101111000001
test_target1/test_source2
test_target1/test_source2/fs2
test_target1/test_source2/fs2/sub
test_target1/test_source2/fs2/sub@test-20101111000000
test_target1/test_source2/fs2/sub@test-20101111000001
""")

        with self.subTest("verify holds"):

            r=shelltest("zfs get -r userrefs test_source1 test_source2 test_target1")
            self.assertMultiLineEqual(r,"""
NAME                                                   PROPERTY  VALUE     SOURCE
test_source1                                           userrefs  -         -
test_source1/fs1                                       userrefs  -         -
test_source1/fs1@test-20101111000000                   userrefs  0         -
test_source1/fs1@test-20101111000001                   userrefs  1         -
test_source1/fs1/sub                                   userrefs  -         -
test_source1/fs1/sub@test-20101111000000               userrefs  0         -
test_source1/fs1/sub@test-20101111000001               userrefs  1         -
test_source2                                           userrefs  -         -
test_source2/fs2                                       userrefs  -         -
test_source2/fs2/sub                                   userrefs  -         -
test_source2/fs2/sub@test-20101111000000               userrefs  0         -
test_source2/fs2/sub@test-20101111000001               userrefs  1         -
test_source2/fs3                                       userrefs  -         -
test_source2/fs3/sub                                   userrefs  -         -
test_target1                                           userrefs  -         -
test_target1/test_source1                              userrefs  -         -
test_target1/test_source1/fs1                          userrefs  -         -
test_target1/test_source1/fs1@test-20101111000000      userrefs  0         -
test_target1/test_source1/fs1@test-20101111000001      userrefs  1         -
test_target1/test_source1/fs1/sub                      userrefs  -         -
test_target1/test_source1/fs1/sub@test-20101111000000  userrefs  0         -
test_target1/test_source1/fs1/sub@test-20101111000001  userrefs  1         -
test_target1/test_source2                              userrefs  -         -
test_target1/test_source2/fs2                          userrefs  -         -
test_target1/test_source2/fs2/sub                      userrefs  -         -
test_target1/test_source2/fs2/sub@test-20101111000000  userrefs  0         -
test_target1/test_source2/fs2/sub@test-20101111000001  userrefs  1         -
""")

        #make sure time handling is correctly. try to make snapshots a year appart and verify that only snapshots mostly 1y old are kept
        #So in this case we only want to see 2 snapshots of 2011, and none of the 2010's anymore.
        with self.subTest("test time checking"):
            with mocktime("20111211000000"):
                self.assertFalse(ZfsAutobackup("test test_target1 --allow-empty --verbose --no-progress".split(" ")).run())

            with mocktime("20111211000001"):
                    self.assertFalse(ZfsAutobackup("test test_target1 --allow-empty --verbose --keep-source 1y1y --keep-target 1d1y --no-progress".split(" ")).run())

            r=shelltest("zfs list -H -o name -r -t all "+TEST_POOLS)
            self.assertMultiLineEqual(r,"""
test_source1
test_source1/fs1
test_source1/fs1@test-20111211000000
test_source1/fs1@test-20111211000001
test_source1/fs1/sub
test_source1/fs1/sub@test-20111211000000
test_source1/fs1/sub@test-20111211000001
test_source2
test_source2/fs2
test_source2/fs2/sub
test_source2/fs2/sub@test-20111211000000
test_source2/fs2/sub@test-20111211000001
test_source2/fs3
test_source2/fs3/sub
test_target1
test_target1/test_source1
test_target1/test_source1/fs1
test_target1/test_source1/fs1@test-20111211000000
test_target1/test_source1/fs1@test-20111211000001
test_target1/test_source1/fs1/sub
test_target1/test_source1/fs1/sub@test-20111211000000
test_target1/test_source1/fs1/sub@test-20111211000001
test_target1/test_source2
test_target1/test_source2/fs2
test_target1/test_source2/fs2/sub
test_target1/test_source2/fs2/sub@test-20111211000000
test_target1/test_source2/fs2/sub@test-20111211000001
""")

    def  test_ignore_othersnaphots(self):

        r=shelltest("zfs snapshot test_source1/fs1@othersimple")
        r=shelltest("zfs snapshot test_source1/fs1@otherdate-20001111000000")

        with mocktime("20101111000000"):
            self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose".split(" ")).run())

            r=shelltest("zfs list -H -o name -r -t all "+TEST_POOLS)
            self.assertMultiLineEqual(r,"""
test_source1
test_source1/fs1
test_source1/fs1@othersimple
test_source1/fs1@otherdate-20001111000000
test_source1/fs1@test-20101111000000
test_source1/fs1/sub
test_source1/fs1/sub@test-20101111000000
test_source2
test_source2/fs2
test_source2/fs2/sub
test_source2/fs2/sub@test-20101111000000
test_source2/fs3
test_source2/fs3/sub
test_target1
test_target1/test_source1
test_target1/test_source1/fs1
test_target1/test_source1/fs1@test-20101111000000
test_target1/test_source1/fs1/sub
test_target1/test_source1/fs1/sub@test-20101111000000
test_target1/test_source2
test_target1/test_source2/fs2
test_target1/test_source2/fs2/sub
test_target1/test_source2/fs2/sub@test-20101111000000
""")

    def  test_othersnaphots(self):

        r=shelltest("zfs snapshot test_source1/fs1@othersimple")
        r=shelltest("zfs snapshot test_source1/fs1@otherdate-20001111000000")

        with mocktime("20101111000000"):
            self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --other-snapshots".split(" ")).run())

            r=shelltest("zfs list -H -o name -r -t all "+TEST_POOLS)
            self.assertMultiLineEqual(r,"""
test_source1
test_source1/fs1
test_source1/fs1@othersimple
test_source1/fs1@otherdate-20001111000000
test_source1/fs1@test-20101111000000
test_source1/fs1/sub
test_source1/fs1/sub@test-20101111000000
test_source2
test_source2/fs2
test_source2/fs2/sub
test_source2/fs2/sub@test-20101111000000
test_source2/fs3
test_source2/fs3/sub
test_target1
test_target1/test_source1
test_target1/test_source1/fs1
test_target1/test_source1/fs1@othersimple
test_target1/test_source1/fs1@otherdate-20001111000000
test_target1/test_source1/fs1@test-20101111000000
test_target1/test_source1/fs1/sub
test_target1/test_source1/fs1/sub@test-20101111000000
test_target1/test_source2
test_target1/test_source2/fs2
test_target1/test_source2/fs2/sub
test_target1/test_source2/fs2/sub@test-20101111000000
""")


    def  test_nosnapshot(self):

        with mocktime("20101111000000"):
            self.assertFalse(ZfsAutobackup("test test_target1 --verbose --no-snapshot --no-progress".split(" ")).run())

            r=shelltest("zfs list -H -o name -r -t all "+TEST_POOLS)
            #(only parents are created )
            #TODO: it probably shouldn't create these
            self.assertMultiLineEqual(r,"""
test_source1
test_source1/fs1
test_source1/fs1/sub
test_source2
test_source2/fs2
test_source2/fs2/sub
test_source2/fs3
test_source2/fs3/sub
test_target1
test_target1/test_source1
test_target1/test_source2
test_target1/test_source2/fs2
""")


    def  test_nosend(self):

        with mocktime("20101111000000"):
            self.assertFalse(ZfsAutobackup("test test_target1 --verbose --no-send --no-progress".split(" ")).run())

            r=shelltest("zfs list -H -o name -r -t all "+TEST_POOLS)
            self.assertMultiLineEqual(r,"""
test_source1
test_source1/fs1
test_source1/fs1@test-20101111000000
test_source1/fs1/sub
test_source1/fs1/sub@test-20101111000000
test_source2
test_source2/fs2
test_source2/fs2/sub
test_source2/fs2/sub@test-20101111000000
test_source2/fs3
test_source2/fs3/sub
test_target1
""")


    def  test_ignorereplicated(self):
        r=shelltest("zfs snapshot test_source1/fs1@otherreplication")

        with mocktime("20101111000000"):
            self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --ignore-replicated".split(" ")).run())

            r=shelltest("zfs list -H -o name -r -t all "+TEST_POOLS)
            self.assertMultiLineEqual(r,"""
test_source1
test_source1/fs1
test_source1/fs1@otherreplication
test_source1/fs1/sub
test_source1/fs1/sub@test-20101111000000
test_source2
test_source2/fs2
test_source2/fs2/sub
test_source2/fs2/sub@test-20101111000000
test_source2/fs3
test_source2/fs3/sub
test_target1
test_target1/test_source1
test_target1/test_source1/fs1
test_target1/test_source1/fs1/sub
test_target1/test_source1/fs1/sub@test-20101111000000
test_target1/test_source2
test_target1/test_source2/fs2
test_target1/test_source2/fs2/sub
test_target1/test_source2/fs2/sub@test-20101111000000
""")

    def  test_noholds(self):

        with mocktime("20101111000000"):
            self.assertFalse(ZfsAutobackup("test test_target1 --verbose --no-holds --no-progress".split(" ")).run())

            r=shelltest("zfs get -r userrefs test_source1 test_source2 test_target1")
            self.assertMultiLineEqual(r,"""
NAME                                                   PROPERTY  VALUE     SOURCE
test_source1                                           userrefs  -         -
test_source1/fs1                                       userrefs  -         -
test_source1/fs1@test-20101111000000                   userrefs  0         -
test_source1/fs1/sub                                   userrefs  -         -
test_source1/fs1/sub@test-20101111000000               userrefs  0         -
test_source2                                           userrefs  -         -
test_source2/fs2                                       userrefs  -         -
test_source2/fs2/sub                                   userrefs  -         -
test_source2/fs2/sub@test-20101111000000               userrefs  0         -
test_source2/fs3                                       userrefs  -         -
test_source2/fs3/sub                                   userrefs  -         -
test_target1                                           userrefs  -         -
test_target1/test_source1                              userrefs  -         -
test_target1/test_source1/fs1                          userrefs  -         -
test_target1/test_source1/fs1@test-20101111000000      userrefs  0         -
test_target1/test_source1/fs1/sub                      userrefs  -         -
test_target1/test_source1/fs1/sub@test-20101111000000  userrefs  0         -
test_target1/test_source2                              userrefs  -         -
test_target1/test_source2/fs2                          userrefs  -         -
test_target1/test_source2/fs2/sub                      userrefs  -         -
test_target1/test_source2/fs2/sub@test-20101111000000  userrefs  0         -
""")


    def  test_strippath(self):

        with mocktime("20101111000000"):
            self.assertFalse(ZfsAutobackup("test test_target1 --verbose --strip-path=1 --no-progress".split(" ")).run())

            r=shelltest("zfs list -H -o name -r -t all "+TEST_POOLS)
            self.assertMultiLineEqual(r,"""
test_source1
test_source1/fs1
test_source1/fs1@test-20101111000000
test_source1/fs1/sub
test_source1/fs1/sub@test-20101111000000
test_source2
test_source2/fs2
test_source2/fs2/sub
test_source2/fs2/sub@test-20101111000000
test_source2/fs3
test_source2/fs3/sub
test_target1
test_target1/fs1
test_target1/fs1@test-20101111000000
test_target1/fs1/sub
test_target1/fs1/sub@test-20101111000000
test_target1/fs2
test_target1/fs2/sub
test_target1/fs2/sub@test-20101111000000
""")

    def test_strippath_collision(self):
        with self.assertRaisesRegexp(Exception,"collision"):
            ZfsAutobackup("test test_target1 --verbose --strip-path=2 --no-progress --debug".split(" ")).run()

    def test_strippath_toomuch(self):
        with self.assertRaisesRegexp(Exception,"too much"):
            ZfsAutobackup("test test_target1 --verbose --strip-path=3 --no-progress --debug".split(" ")).run()

    def  test_clearrefres(self):

        #on zfs utils 0.6.x -x isnt supported
        r=shelltest("zfs recv -x bla test >/dev/null </dev/zero; echo $?")
        if r=="\n2\n":
            self.skipTest("This zfs-userspace version doesnt support -x")

        r=shelltest("zfs set refreservation=1M test_source1/fs1")

        with mocktime("20101111000000"):
            self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --clear-refreservation".split(" ")).run())

            r=shelltest("zfs get -r refreservation test_source1 test_source2 test_target1")
            self.assertMultiLineEqual(r,"""
NAME                                                   PROPERTY        VALUE      SOURCE
test_source1                                           refreservation  none       default
test_source1/fs1                                       refreservation  1M         local
test_source1/fs1@test-20101111000000                   refreservation  -          -
test_source1/fs1/sub                                   refreservation  none       default
test_source1/fs1/sub@test-20101111000000               refreservation  -          -
test_source2                                           refreservation  none       default
test_source2/fs2                                       refreservation  none       default
test_source2/fs2/sub                                   refreservation  none       default
test_source2/fs2/sub@test-20101111000000               refreservation  -          -
test_source2/fs3                                       refreservation  none       default
test_source2/fs3/sub                                   refreservation  none       default
test_target1                                           refreservation  none       default
test_target1/test_source1                              refreservation  none       default
test_target1/test_source1/fs1                          refreservation  none       default
test_target1/test_source1/fs1@test-20101111000000      refreservation  -          -
test_target1/test_source1/fs1/sub                      refreservation  none       default
test_target1/test_source1/fs1/sub@test-20101111000000  refreservation  -          -
test_target1/test_source2                              refreservation  none       default
test_target1/test_source2/fs2                          refreservation  none       default
test_target1/test_source2/fs2/sub                      refreservation  none       default
test_target1/test_source2/fs2/sub@test-20101111000000  refreservation  -          -
""")


    def  test_clearmount(self):

        #on zfs utils 0.6.x -o isnt supported
        r=shelltest("zfs recv -o bla=1 test >/dev/null </dev/zero; echo $?")
        if r=="\n2\n":
            self.skipTest("This zfs-userspace version doesnt support -o")


        with mocktime("20101111000000"):
            self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --clear-mountpoint --debug".split(" ")).run())

            r=shelltest("zfs get -r canmount test_source1 test_source2 test_target1")
            self.assertMultiLineEqual(r,"""
NAME                                                   PROPERTY  VALUE     SOURCE
test_source1                                           canmount  on        default
test_source1/fs1                                       canmount  on        default
test_source1/fs1@test-20101111000000                   canmount  -         -
test_source1/fs1/sub                                   canmount  on        default
test_source1/fs1/sub@test-20101111000000               canmount  -         -
test_source2                                           canmount  on        default
test_source2/fs2                                       canmount  on        default
test_source2/fs2/sub                                   canmount  on        default
test_source2/fs2/sub@test-20101111000000               canmount  -         -
test_source2/fs3                                       canmount  on        default
test_source2/fs3/sub                                   canmount  on        default
test_target1                                           canmount  on        default
test_target1/test_source1                              canmount  off       local
test_target1/test_source1/fs1                          canmount  noauto    local
test_target1/test_source1/fs1@test-20101111000000      canmount  -         -
test_target1/test_source1/fs1/sub                      canmount  noauto    local
test_target1/test_source1/fs1/sub@test-20101111000000  canmount  -         -
test_target1/test_source2                              canmount  off       local
test_target1/test_source2/fs2                          canmount  off       local
test_target1/test_source2/fs2/sub                      canmount  noauto    local
test_target1/test_source2/fs2/sub@test-20101111000000  canmount  -         -
""")


    def  test_rollback(self):

        #initial backup
        with mocktime("20101111000000"):
            self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose".split(" ")).run())

        #make change
        r=shelltest("touch /test_target1/test_source1/fs1/change.txt")

        with mocktime("20101111000001"):
            #should fail (busy)
            self.assertTrue(ZfsAutobackup("test test_target1 --no-progress --verbose --allow-empty".split(" ")).run())

        with mocktime("20101111000002"):
            #rollback, should succeed
            self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --allow-empty --rollback".split(" ")).run())


    def  test_destroyincompat(self):

        #initial backup
        with mocktime("20101111000000"):
            self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose".split(" ")).run())

        #add multiple compatible snapshot (written is still 0)
        r=shelltest("zfs snapshot test_target1/test_source1/fs1@compatible1")
        r=shelltest("zfs snapshot test_target1/test_source1/fs1@compatible2")

        with mocktime("20101111000001"):
            #should be ok, is compatible
            self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --allow-empty".split(" ")).run())

        #add incompatible snapshot by changing and snapshotting
        r=shelltest("touch /test_target1/test_source1/fs1/change.txt")
        r=shelltest("zfs snapshot test_target1/test_source1/fs1@incompatible1")


        with mocktime("20101111000002"):
            #--test should fail, now incompatible
            self.assertTrue(ZfsAutobackup("test test_target1 --no-progress --verbose --allow-empty --test".split(" ")).run())

        with mocktime("20101111000002"):
            #should fail, now incompatible
            self.assertTrue(ZfsAutobackup("test test_target1 --no-progress --verbose --allow-empty".split(" ")).run())

        with mocktime("20101111000003"):
            #--test should succeed by destroying incompatibles
            self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --allow-empty --destroy-incompatible --test".split(" ")).run())

        with mocktime("20101111000003"):
            #should succeed by destroying incompatibles
            self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --allow-empty --destroy-incompatible".split(" ")).run())

        r = shelltest("zfs list -H -o name -r -t all test_target1")
        self.assertMultiLineEqual(r, """
test_target1
test_target1/test_source1
test_target1/test_source1/fs1
test_target1/test_source1/fs1@test-20101111000000
test_target1/test_source1/fs1@compatible1
test_target1/test_source1/fs1@compatible2
test_target1/test_source1/fs1@test-20101111000001
test_target1/test_source1/fs1@test-20101111000002
test_target1/test_source1/fs1@test-20101111000003
test_target1/test_source1/fs1/sub
test_target1/test_source1/fs1/sub@test-20101111000000
test_target1/test_source1/fs1/sub@test-20101111000001
test_target1/test_source1/fs1/sub@test-20101111000002
test_target1/test_source1/fs1/sub@test-20101111000003
test_target1/test_source2
test_target1/test_source2/fs2
test_target1/test_source2/fs2/sub
test_target1/test_source2/fs2/sub@test-20101111000000
test_target1/test_source2/fs2/sub@test-20101111000001
test_target1/test_source2/fs2/sub@test-20101111000002
test_target1/test_source2/fs2/sub@test-20101111000003
""")






    def test_ssh(self):

        #test all ssh directions

        with mocktime("20101111000000"):
            self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --allow-empty --ssh-source localhost --exclude-received".split(" ")).run())

        with mocktime("20101111000001"):
            self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --allow-empty --ssh-target localhost --exclude-received".split(" ")).run())

        with mocktime("20101111000002"):
            self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --allow-empty --ssh-source localhost --ssh-target localhost".split(" ")).run())


        r=shelltest("zfs list -H -o name -r -t all "+TEST_POOLS)
        self.assertMultiLineEqual(r,"""
test_source1
test_source1/fs1
test_source1/fs1@test-20101111000000
test_source1/fs1@test-20101111000001
test_source1/fs1@test-20101111000002
test_source1/fs1/sub
test_source1/fs1/sub@test-20101111000000
test_source1/fs1/sub@test-20101111000001
test_source1/fs1/sub@test-20101111000002
test_source2
test_source2/fs2
test_source2/fs2/sub
test_source2/fs2/sub@test-20101111000000
test_source2/fs2/sub@test-20101111000001
test_source2/fs2/sub@test-20101111000002
test_source2/fs3
test_source2/fs3/sub
test_target1
test_target1/test_source1
test_target1/test_source1/fs1
test_target1/test_source1/fs1@test-20101111000000
test_target1/test_source1/fs1@test-20101111000001
test_target1/test_source1/fs1@test-20101111000002
test_target1/test_source1/fs1/sub
test_target1/test_source1/fs1/sub@test-20101111000000
test_target1/test_source1/fs1/sub@test-20101111000001
test_target1/test_source1/fs1/sub@test-20101111000002
test_target1/test_source2
test_target1/test_source2/fs2
test_target1/test_source2/fs2/sub
test_target1/test_source2/fs2/sub@test-20101111000000
test_target1/test_source2/fs2/sub@test-20101111000001
test_target1/test_source2/fs2/sub@test-20101111000002
""")


    def  test_minchange(self):

        #initial
        with mocktime("20101111000000"):
            self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --min-change 100000".split(" ")).run())

        #make small change, use umount to reflect the changes immediately
        r=shelltest("zfs set compress=off test_source1")
        r=shelltest("touch /test_source1/fs1/change.txt")
        r=shelltest("zfs umount test_source1/fs1; zfs mount test_source1/fs1")


        #too small change, takes no snapshots
        with mocktime("20101111000001"):
            self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --min-change 100000".split(" ")).run())

        #make big change
        r=shelltest("dd if=/dev/zero of=/test_source1/fs1/change.txt bs=200000 count=1")
        r=shelltest("zfs umount test_source1/fs1; zfs mount test_source1/fs1")

        #bigger change, should take snapshot
        with mocktime("20101111000002"):
            self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --min-change 100000".split(" ")).run())

        r=shelltest("zfs list -H -o name -r -t all "+TEST_POOLS)
        self.assertMultiLineEqual(r,"""
test_source1
test_source1/fs1
test_source1/fs1@test-20101111000000
test_source1/fs1@test-20101111000002
test_source1/fs1/sub
test_source1/fs1/sub@test-20101111000000
test_source2
test_source2/fs2
test_source2/fs2/sub
test_source2/fs2/sub@test-20101111000000
test_source2/fs3
test_source2/fs3/sub
test_target1
test_target1/test_source1
test_target1/test_source1/fs1
test_target1/test_source1/fs1@test-20101111000000
test_target1/test_source1/fs1@test-20101111000002
test_target1/test_source1/fs1/sub
test_target1/test_source1/fs1/sub@test-20101111000000
test_target1/test_source2
test_target1/test_source2/fs2
test_target1/test_source2/fs2/sub
test_target1/test_source2/fs2/sub@test-20101111000000
""")

    def  test_test(self):

        #initial
        with mocktime("20101111000000"):
            self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --test".split(" ")).run())

        r=shelltest("zfs list -H -o name -r -t all "+TEST_POOLS)
        self.assertMultiLineEqual(r,"""
test_source1
test_source1/fs1
test_source1/fs1/sub
test_source2
test_source2/fs2
test_source2/fs2/sub
test_source2/fs3
test_source2/fs3/sub
test_target1
""")

        #actual make initial backup
        with mocktime("20101111000001"):
            self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose".split(" ")).run())


        #test incremental
        with mocktime("20101111000002"):
            self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --allow-empty --verbose --test".split(" ")).run())

        r=shelltest("zfs list -H -o name -r -t all "+TEST_POOLS)
        self.assertMultiLineEqual(r,"""
test_source1
test_source1/fs1
test_source1/fs1@test-20101111000001
test_source1/fs1/sub
test_source1/fs1/sub@test-20101111000001
test_source2
test_source2/fs2
test_source2/fs2/sub
test_source2/fs2/sub@test-20101111000001
test_source2/fs3
test_source2/fs3/sub
test_target1
test_target1/test_source1
test_target1/test_source1/fs1
test_target1/test_source1/fs1@test-20101111000001
test_target1/test_source1/fs1/sub
test_target1/test_source1/fs1/sub@test-20101111000001
test_target1/test_source2
test_target1/test_source2/fs2
test_target1/test_source2/fs2/sub
test_target1/test_source2/fs2/sub@test-20101111000001
""")


    def test_migrate(self):
        """test migration from other snapshotting systems. zfs-autobackup should be able to continue from any common snapshot, not just its own."""

        shelltest("zfs snapshot test_source1/fs1@migrate1")
        shelltest("zfs create test_target1/test_source1")
        shelltest("zfs send  test_source1/fs1@migrate1| zfs recv test_target1/test_source1/fs1")

        with mocktime("20101111000000"):
            self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose".split(" ")).run())

        r=shelltest("zfs list -H -o name -r -t all "+TEST_POOLS)
        self.assertMultiLineEqual(r,"""
test_source1
test_source1/fs1
test_source1/fs1@migrate1
test_source1/fs1@test-20101111000000
test_source1/fs1/sub
test_source1/fs1/sub@test-20101111000000
test_source2
test_source2/fs2
test_source2/fs2/sub
test_source2/fs2/sub@test-20101111000000
test_source2/fs3
test_source2/fs3/sub
test_target1
test_target1/test_source1
test_target1/test_source1/fs1
test_target1/test_source1/fs1@migrate1
test_target1/test_source1/fs1@test-20101111000000
test_target1/test_source1/fs1/sub
test_target1/test_source1/fs1/sub@test-20101111000000
test_target1/test_source2
test_target1/test_source2/fs2
test_target1/test_source2/fs2/sub
test_target1/test_source2/fs2/sub@test-20101111000000
""")

    def test_keep0(self):
        """test if keep-source=0 and keep-target=0 dont delete common snapshot and break backup"""

        with mocktime("20101111000000"):
            self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --keep-source=0 --keep-target=0".split(" ")).run())

        #make snapshot, shouldnt delete 0
        with mocktime("20101111000001"):
            self.assertFalse(ZfsAutobackup("test --no-progress --verbose --keep-source=0 --keep-target=0 --allow-empty".split(" ")).run())

        #make snapshot 2, shouldnt delete 0 since it has holds, but will delete 1 since it has no holds
        with mocktime("20101111000002"):
            self.assertFalse(ZfsAutobackup("test --no-progress --verbose --keep-source=0 --keep-target=0 --allow-empty".split(" ")).run())

        r = shelltest("zfs list -H -o name -r -t all " + TEST_POOLS)
        self.assertMultiLineEqual(r, """
test_source1
test_source1/fs1
test_source1/fs1@test-20101111000000
test_source1/fs1@test-20101111000002
test_source1/fs1/sub
test_source1/fs1/sub@test-20101111000000
test_source1/fs1/sub@test-20101111000002
test_source2
test_source2/fs2
test_source2/fs2/sub
test_source2/fs2/sub@test-20101111000000
test_source2/fs2/sub@test-20101111000002
test_source2/fs3
test_source2/fs3/sub
test_target1
test_target1/test_source1
test_target1/test_source1/fs1
test_target1/test_source1/fs1@test-20101111000000
test_target1/test_source1/fs1/sub
test_target1/test_source1/fs1/sub@test-20101111000000
test_target1/test_source2
test_target1/test_source2/fs2
test_target1/test_source2/fs2/sub
test_target1/test_source2/fs2/sub@test-20101111000000
""")

        #make another backup but with no-holds. we should naturally endup with only number 3
        with mocktime("20101111000003"):
            self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --keep-source=0 --keep-target=0 --no-holds --allow-empty".split(" ")).run())

        r = shelltest("zfs list -H -o name -r -t all " + TEST_POOLS)
        self.assertMultiLineEqual(r, """
test_source1
test_source1/fs1
test_source1/fs1@test-20101111000003
test_source1/fs1/sub
test_source1/fs1/sub@test-20101111000003
test_source2
test_source2/fs2
test_source2/fs2/sub
test_source2/fs2/sub@test-20101111000003
test_source2/fs3
test_source2/fs3/sub
test_target1
test_target1/test_source1
test_target1/test_source1/fs1
test_target1/test_source1/fs1@test-20101111000003
test_target1/test_source1/fs1/sub
test_target1/test_source1/fs1/sub@test-20101111000003
test_target1/test_source2
test_target1/test_source2/fs2
test_target1/test_source2/fs2/sub
test_target1/test_source2/fs2/sub@test-20101111000003
""")


        # run with snapshot-only for 4, since we used no-holds, it will delete 3 on the source, breaking the backup
        with mocktime("20101111000004"):
            self.assertFalse(ZfsAutobackup("test --no-progress --verbose --keep-source=0 --keep-target=0 --allow-empty".split(" ")).run())

        r = shelltest("zfs list -H -o name -r -t all " + TEST_POOLS)
        self.assertMultiLineEqual(r, """
test_source1
test_source1/fs1
test_source1/fs1@test-20101111000004
test_source1/fs1/sub
test_source1/fs1/sub@test-20101111000004
test_source2
test_source2/fs2
test_source2/fs2/sub
test_source2/fs2/sub@test-20101111000004
test_source2/fs3
test_source2/fs3/sub
test_target1
test_target1/test_source1
test_target1/test_source1/fs1
test_target1/test_source1/fs1@test-20101111000003
test_target1/test_source1/fs1/sub
test_target1/test_source1/fs1/sub@test-20101111000003
test_target1/test_source2
test_target1/test_source2/fs2
test_target1/test_source2/fs2/sub
test_target1/test_source2/fs2/sub@test-20101111000003
""")


    def test_progress(self):

        r=shelltest("dd if=/dev/urandom of=/test_source1/data.txt bs=5M count=1")
        r = shelltest("zfs snapshot test_source1@test")

        l=LogConsole(show_verbose=True, show_debug=True, color=False)
        n=ZfsNode(utc=False, snapshot_time_format="bla", hold_name="bla", logger=l)
        d=ZfsDataset(n,"test_source1@test")

        sp=d.send_pipe([], prev_snapshot=None, resume_token=None, show_progress=True, raw=False, send_pipes=[], send_properties=True, write_embedded=True, zfs_compressed=True)




        with OutputIO() as buf:
            with redirect_stderr(buf):
                try:

                    p=n.run(["mbuffer", "-R1M", "-m4096", "-o" ,"/dev/null"], inp=sp)
                    # p=n.run(["dd", "of=/dev/null"], inp=sp)

                except:
                    pass

            print(list(buf.getvalue()))
            # correct message?
            self.assertRegex(buf.getvalue(),".*>>> .*minutes left.*")



# This is the low level process executing stuff.
# It makes piping and parallel process handling more easy.

# You can specify a handler for each line of stderr output for each item in the pipe.
# Every item also has its own exitcode handler.

# Normally you add a stdout_handler to the last item in the pipe.
# However: You can also add stdout_handler to other items in a pipe. This will turn that item in to a manual pipe: your
# handler is responsible for sending data into the next item of the pipe. (avaiable in item.next)

# You can also use manual pipe mode to just execute multiple command in parallel and handle their output parallel,
# without doing any actual pipe stuff. (because you dont HAVE to send data into the next item.)


import subprocess
import os
import select

try:
    from shlex import quote as cmd_quote
except ImportError:
    from pipes import quote as cmd_quote


class CmdItem:
    """one command item, to be added to a CmdPipe"""

    def __init__(self, cmd, readonly=False, stderr_handler=None, exit_handler=None, stdout_handler=None, shell=False):
        """create item. caller has to make sure cmd is properly escaped when using shell.

        If stdout_handler is None, it will connect the stdout to the stdin of the next item in the pipe, like
        and actual system pipe. (no python overhead)

        :type cmd: list of str
        """

        self.cmd = cmd
        self.readonly = readonly
        self.stderr_handler = stderr_handler
        self.stdout_handler = stdout_handler
        self.exit_handler = exit_handler
        self.shell = shell
        self.process = None
        self.next = None #next item in pipe, set by CmdPipe

    def __str__(self):
        """return copy-pastable version of command."""
        if self.shell:
            # its already copy pastable for a shell:
            return " ".join(self.cmd)
        else:
            # make it copy-pastable, will make a mess of quotes sometimes, but is correct
            return " ".join(map(cmd_quote, self.cmd))

    def create(self, stdin):
        """actually create the subprocess (called by CmdPipe)"""

        # make sure the command gets all the data in utf8 format:
        # (this is necessary if LC_ALL=en_US.utf8 is not set in the environment)
        encoded_cmd = []
        for arg in self.cmd:
            encoded_cmd.append(arg.encode('utf-8'))

        self.process = subprocess.Popen(encoded_cmd, env=os.environ, stdout=subprocess.PIPE, stdin=stdin,
                                        stderr=subprocess.PIPE, shell=self.shell)


class CmdPipe:
    """a pipe of one or more commands. also takes care of utf-8 encoding/decoding and line based parsing"""

    def __init__(self, readonly=False, inp=None):
        """
        :param inp: input string for stdin
        :param readonly: Only execute if entire pipe consist of readonly commands
        """
        # list of commands + error handlers to execute
        self.items = []

        self.inp = inp
        self.readonly = readonly
        self._should_execute = True

    def add(self, cmd_item):
        """adds a CmdItem to pipe.
        :type cmd_item: CmdItem
        """

        self.items.append(cmd_item)

        if not cmd_item.readonly and self.readonly:
            self._should_execute = False

    def __str__(self):
        """transform whole pipe into oneliner for debugging and testing. this should generate a copy-pastable string for in a console """

        ret = ""
        for item in self.items:
            if ret:
                ret = ret + " | "
            ret = ret + "({})".format(item)  # this will do proper escaping to make it copypastable

        return ret

    def should_execute(self):
        return self._should_execute

    def execute(self):
        """run the pipe. returns True all exit handlers returned true. (otherwise it will be False/None depending on exit handlers returncode) """

        if not self._should_execute:
            return True

        selectors = self.__create()

        if not selectors:
            raise (Exception("Cant use cmdpipe without any output handlers."))

        self.__process_outputs(selectors)

        # close filehandles
        for item in self.items:
            item.process.stderr.close()
            item.process.stdout.close()

        # call exit handlers
        success = True
        for item in self.items:
            if item.exit_handler is not None:
                success=item.exit_handler(item.process.returncode) and success

        return success

    def __process_outputs(self, selectors):
        """watch all output selectors and call handlers"""

        while True:
            # wait for output on one of the stderrs or last_stdout
            (read_ready, write_ready, ex_ready) = select.select(selectors, [], [])

            eof_count = 0
            done_count = 0

            # read line and call appropriate handlers

            for item in self.items:
                if item.process.stdout in read_ready:
                    line = item.process.stdout.readline().decode('utf-8').rstrip()
                    if line != "":
                        item.stdout_handler(line)
                    else:
                        eof_count = eof_count + 1
                        if item.next:
                            item.next.process.stdin.close()

                if item.process.stderr in read_ready:
                    line = item.process.stderr.readline().decode('utf-8').rstrip()
                    if line != "":
                        item.stderr_handler(line)
                    else:
                        eof_count = eof_count + 1


                if item.process.poll() is not None:
                    done_count = done_count + 1

            # all filehandles are eof and all processes are done (poll() is not None)
            if eof_count == len(selectors) and done_count == len(self.items):
                break



    def __create(self):
        """create actual processes, do piping and return selectors."""

        selectors = []
        next_stdin = subprocess.PIPE  # means we write input via python instead of an actual system pipe
        first = True
        prev_item = None

        for item in self.items:

            # creates the actual subprocess via subprocess.popen
            item.create(next_stdin)

            # we piped previous process? dont forget to close its stdout
            if next_stdin != subprocess.PIPE:
                next_stdin.close()

            if item.stderr_handler:
                selectors.append(item.process.stderr)

            # we're the first process in the pipe
            if first:
                if self.inp is not None:
                    # write the input we have
                    item.process.stdin.write(self.inp.encode('utf-8'))
                item.process.stdin.close()
                first = False

            # manual stdout handling or pipe it to the next process?
            if item.stdout_handler is None:
                # no manual stdout handling, pipe it to the next process via sytem pipe
                next_stdin = item.process.stdout
            else:
                # manual stdout handling via python
                selectors.append(item.process.stdout)
                # next process will get input from python:
                next_stdin = subprocess.PIPE

            if prev_item is not None:
                prev_item.next = item

            prev_item = item
        return selectors




# NOTE: surprisingly sha1 in via python3 is faster than the native sha1sum utility, even in the way we use below!
import os
import platform
import sys
from datetime import datetime


def tmp_name(suffix=""):
    """create temporary name unique to this process and node. always retruns the same result during the same execution"""

    #we could use uuids but those are ugly and confusing
    name="{}-{}-{}".format(
        os.path.basename(sys.argv[0]).replace(" ","_"),
        platform.node(),
        os.getpid())
    name=name+suffix
    return name


def get_tmp_clone_name(snapshot):
    pool=snapshot.zfs_node.get_pool(snapshot)
    return pool.name+"/"+tmp_name()



def output_redir():
    """use this after a BrokenPipeError to prevent further exceptions.
    Redirects stdout/err to /dev/null
    """

    devnull = os.open(os.devnull, os.O_WRONLY)
    os.dup2(devnull, sys.stdout.fileno())
    os.dup2(devnull, sys.stderr.fileno())

def sigpipe_handler(sig, stack):
    #redir output so we dont get more SIGPIPES during cleanup. (which my try to write to stdout)
    output_redir()
    #deb('redir')

# def check_output():
#     """make sure stdout still functions. if its broken, this will trigger a SIGPIPE which will be handled by the sigpipe_handler."""
#     try:
#         print(" ")
#         sys.stdout.flush()
#     except Exception as e:
#         pass

# def deb(txt):
#     with open('/tmp/debug.log', 'a') as fh:
#         fh.write("DEB: "+txt+"\n")


# This should be the only source of trueth for the current datetime.
# This function will be mocked during unit testing.


datetime_now_mock=None
def datetime_now(utc):
    if datetime_now_mock is None:
        return( datetime.utcnow() if utc else datetime.now())
    else:
        return datetime_now_mock



# NOTE: this should inherit from (object) to function correctly with python 2.7
class CachedProperty(object):
    """ A property that is only computed once per instance and
    then stores the result in _cached_properties of the object.

        Source: https://github.com/bottlepy/bottle/commit/fa7733e075da0d790d809aa3d2f53071897e6f76
        """

    def __init__(self, func):
        self.__doc__ = getattr(func, '__doc__')
        self.func = func

    def __get__(self, obj, cls):
        if obj is None:
            return self

        propname = self.func.__name__

        if not hasattr(obj, '_cached_properties'):
            obj._cached_properties = {}

        if propname not in obj._cached_properties:
            obj._cached_properties[propname] = self.func(obj)
            # value = obj.__dict__[propname] = self.func(obj)

        return obj._cached_properties[propname]

    @staticmethod
    def clear(obj):
        """clears cache of obj"""
        if hasattr(obj, '_cached_properties'):
            obj._cached_properties = {}

    @staticmethod
    def is_cached(obj, propname):
        if hasattr(obj, '_cached_properties') and propname in obj._cached_properties:
            return True
        else:
            return False


from .CachedProperty import CachedProperty


class ZfsPool():
    """a zfs pool"""

    def __init__(self, zfs_node, name):
        """name: name of the pool
        """

        self.zfs_node = zfs_node
        self.name = name

    def __repr__(self):
        return "{}: {}".format(self.zfs_node, self.name)

    def __str__(self):
        return self.name

    def __eq__(self, obj):
        if not isinstance(obj, ZfsPool):
            return False

        return self.name == obj.name

    def verbose(self, txt):
        self.zfs_node.verbose("zpool {}: {}".format(self.name, txt))

    def error(self, txt):
        self.zfs_node.error("zpool {}: {}".format(self.name, txt))

    def debug(self, txt):
        self.zfs_node.debug("zpool {}: {}".format(self.name, txt))

    @CachedProperty
    def properties(self):
        """all zpool properties"""

        self.debug("Getting zpool properties")

        cmd = [
            "zpool", "get", "-H", "-p", "all", self.name
        ]

        ret = {}

        for pair in self.zfs_node.run(tab_split=True, cmd=cmd, readonly=True, valid_exitcodes=[0]):
            ret[pair[1]] = pair[2]

        return ret

    @property
    def features(self):
        """get list of active zpool features"""

        ret = []
        for (key, value) in self.properties.items():
            if key.startswith("feature@"):
                feature = key.split("@")[1]
                if value == 'enabled' or value == 'active':
                    ret.append(feature)

        return ret


import zfs_autobackup.compressors
from basetest import *
import time


class TestSendRecvPipes(unittest2.TestCase):
    """test input/output pipes for zfs send and recv"""

    def setUp(self):
        prepare_zpools()
        self.longMessage = True

    def test_send_basics(self):
        """send basics (remote/local send pipe)"""

        with self.subTest("local local pipe"):
            with mocktime("20101111000000"):
                self.assertFalse(ZfsAutobackup(
                    ["test", "test_target1", "--allow-empty", "--exclude-received", "--no-holds", "--no-progress", "--clear-mountpoint",
                     "--send-pipe=dd bs=1M", "--recv-pipe=dd bs=2M"]).run())

            shelltest("zfs destroy -r test_target1/test_source1/fs1/sub")

        with self.subTest("remote local pipe"):
            with mocktime("20101111000001"):
                self.assertFalse(ZfsAutobackup(
                    ["test", "test_target1", "--allow-empty", "--exclude-received", "--no-holds", "--no-progress",
                     "--ssh-source=localhost", "--send-pipe=dd bs=1M", "--recv-pipe=dd bs=2M"]).run())

            shelltest("zfs destroy -r test_target1/test_source1/fs1/sub")

        with self.subTest("local remote pipe"):
            with mocktime("20101111000002"):
                self.assertFalse(ZfsAutobackup(
                    ["test", "test_target1", "--allow-empty", "--exclude-received", "--no-holds", "--no-progress",
                     "--ssh-target=localhost", "--send-pipe=dd bs=1M", "--recv-pipe=dd bs=2M"]).run())

            shelltest("zfs destroy -r test_target1/test_source1/fs1/sub")

        with self.subTest("remote remote pipe"):
            with mocktime("20101111000003"):
                self.assertFalse(ZfsAutobackup(
                    ["test", "test_target1", "--allow-empty", "--exclude-received", "--no-holds", "--no-progress",
                     "--ssh-source=localhost", "--ssh-target=localhost", "--send-pipe=dd bs=1M",
                     "--recv-pipe=dd bs=2M"]).run())

        r = shelltest("zfs list -H -o name -r -t all test_target1")
        self.assertMultiLineEqual(r, """
test_target1
test_target1/test_source1
test_target1/test_source1/fs1
test_target1/test_source1/fs1@test-20101111000000
test_target1/test_source1/fs1@test-20101111000001
test_target1/test_source1/fs1@test-20101111000002
test_target1/test_source1/fs1@test-20101111000003
test_target1/test_source1/fs1/sub
test_target1/test_source1/fs1/sub@test-20101111000000
test_target1/test_source1/fs1/sub@test-20101111000001
test_target1/test_source1/fs1/sub@test-20101111000002
test_target1/test_source1/fs1/sub@test-20101111000003
test_target1/test_source2
test_target1/test_source2/fs2
test_target1/test_source2/fs2/sub
test_target1/test_source2/fs2/sub@test-20101111000000
test_target1/test_source2/fs2/sub@test-20101111000001
test_target1/test_source2/fs2/sub@test-20101111000002
test_target1/test_source2/fs2/sub@test-20101111000003
""")

    def test_compress(self):
        """send basics (remote/local send pipe)"""

        for compress in zfs_autobackup.compressors.COMPRESS_CMDS.keys():
            with self.subTest("compress " + compress):
                with mocktime("20101111000000"):
                    self.assertFalse(ZfsAutobackup(
                        ["test", "test_target1", "--exclude-received", "--no-holds", "--no-progress", "--verbose",
                         "--compress=" + compress]).run())

                shelltest("zfs destroy -r test_target1/test_source1/fs1/sub")

    def test_buffer(self):
        """test different buffer configurations"""

        with self.subTest("local local pipe"):
            with mocktime("20101111000000"):
                self.assertFalse(ZfsAutobackup(
                    ["test", "test_target1", "--allow-empty", "--exclude-received", "--no-holds", "--no-progress", "--clear-mountpoint", "--buffer=1M"]).run())

            shelltest("zfs destroy -r test_target1/test_source1/fs1/sub")

        with self.subTest("remote local pipe"):
            with mocktime("20101111000001"):
                self.assertFalse(ZfsAutobackup(
                    ["test", "test_target1", "--allow-empty", "--verbose", "--exclude-received", "--no-holds",
                     "--no-progress", "--ssh-source=localhost", "--buffer=1M"]).run())

            shelltest("zfs destroy -r test_target1/test_source1/fs1/sub")

        with self.subTest("local remote pipe"):
            with mocktime("20101111000002"):
                self.assertFalse(ZfsAutobackup(
                    ["test", "test_target1", "--allow-empty", "--exclude-received", "--no-holds", "--no-progress",
                     "--ssh-target=localhost", "--buffer=1M"]).run())

            shelltest("zfs destroy -r test_target1/test_source1/fs1/sub")

        with self.subTest("remote remote pipe"):
            with mocktime("20101111000003"):
                self.assertFalse(ZfsAutobackup(
                    ["test", "test_target1", "--allow-empty", "--exclude-received", "--no-holds", "--no-progress",
                     "--ssh-source=localhost", "--ssh-target=localhost", "--buffer=1M"]).run())

        r = shelltest("zfs list -H -o name -r -t all test_target1")
        self.assertMultiLineEqual(r, """
test_target1
test_target1/test_source1
test_target1/test_source1/fs1
test_target1/test_source1/fs1@test-20101111000000
test_target1/test_source1/fs1@test-20101111000001
test_target1/test_source1/fs1@test-20101111000002
test_target1/test_source1/fs1@test-20101111000003
test_target1/test_source1/fs1/sub
test_target1/test_source1/fs1/sub@test-20101111000000
test_target1/test_source1/fs1/sub@test-20101111000001
test_target1/test_source1/fs1/sub@test-20101111000002
test_target1/test_source1/fs1/sub@test-20101111000003
test_target1/test_source2
test_target1/test_source2/fs2
test_target1/test_source2/fs2/sub
test_target1/test_source2/fs2/sub@test-20101111000000
test_target1/test_source2/fs2/sub@test-20101111000001
test_target1/test_source2/fs2/sub@test-20101111000002
test_target1/test_source2/fs2/sub@test-20101111000003
""")

    def test_rate(self):
        """test rate limit"""

        start = time.time()
        with mocktime("20101111000000"):
            self.assertFalse(ZfsAutobackup(
                ["test", "test_target1", "--exclude-received", "--no-holds", "--no-progress", "--rate=50k"]).run())

        # not a great way of verifying but it works.
        self.assertGreater(time.time() - start, 5)



from basetest import *

class TestZfsAutobackup32(unittest2.TestCase):
    """various new 3.2 features"""

    def setUp(self):
        prepare_zpools()
        self.longMessage=True

    def test_invalid_common_snapshot(self):

        with mocktime("20101111000000"):
            self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --allow-empty".split(" ")).run())

        #create 2 snapshots with the same name, which are invalid as common snapshot
        shelltest("zfs snapshot test_source1/fs1@invalid")
        shelltest("zfs snapshot test_target1/test_source1/fs1@invalid")

        with mocktime("20101111000001"):
            #try the old way (without guid checking), and fail:
            self.assertEqual(ZfsAutobackup("test test_target1 --no-progress --verbose --allow-empty --no-guid-check".split(" ")).run(),1)
            #new way should be ok:
            self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --no-snapshot".split(" ")).run())

            r=shelltest("zfs list -H -o name -r -t all "+TEST_POOLS)
            self.assertMultiLineEqual(r,"""
test_source1
test_source1/fs1
test_source1/fs1@test-20101111000000
test_source1/fs1@invalid
test_source1/fs1@test-20101111000001
test_source1/fs1/sub
test_source1/fs1/sub@test-20101111000000
test_source1/fs1/sub@test-20101111000001
test_source2
test_source2/fs2
test_source2/fs2/sub
test_source2/fs2/sub@test-20101111000000
test_source2/fs2/sub@test-20101111000001
test_source2/fs3
test_source2/fs3/sub
test_target1
test_target1/test_source1
test_target1/test_source1/fs1
test_target1/test_source1/fs1@test-20101111000000
test_target1/test_source1/fs1@invalid
test_target1/test_source1/fs1@test-20101111000001
test_target1/test_source1/fs1/sub
test_target1/test_source1/fs1/sub@test-20101111000000
test_target1/test_source1/fs1/sub@test-20101111000001
test_target1/test_source2
test_target1/test_source2/fs2
test_target1/test_source2/fs2/sub
test_target1/test_source2/fs2/sub@test-20101111000000
test_target1/test_source2/fs2/sub@test-20101111000001
""")

    def test_invalid_common_snapshot_with_data(self):

        with mocktime("20101111000000"):
            self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --allow-empty".split(" ")).run())

        #create 2 snapshots with the same name, which are invalid as common snapshot
        shelltest("zfs snapshot test_source1/fs1@invalid")
        shelltest("touch /test_target1/test_source1/fs1/shouldnotbeHere")
        shelltest("zfs snapshot test_target1/test_source1/fs1@invalid")

        with mocktime("20101111000001"):
            #try the old way and fail:
            self.assertEqual(ZfsAutobackup("test test_target1 --no-progress --verbose --allow-empty --destroy-incompatible --no-guid-check".split(" ")).run(),1)
            #new way should be ok
            self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --no-snapshot --destroy-incompatible".split(" ")).run())

            r=shelltest("zfs list -H -o name -r -t all "+TEST_POOLS)
            self.assertMultiLineEqual(r,"""
test_source1
test_source1/fs1
test_source1/fs1@test-20101111000000
test_source1/fs1@invalid
test_source1/fs1@test-20101111000001
test_source1/fs1/sub
test_source1/fs1/sub@test-20101111000000
test_source1/fs1/sub@test-20101111000001
test_source2
test_source2/fs2
test_source2/fs2/sub
test_source2/fs2/sub@test-20101111000000
test_source2/fs2/sub@test-20101111000001
test_source2/fs3
test_source2/fs3/sub
test_target1
test_target1/test_source1
test_target1/test_source1/fs1
test_target1/test_source1/fs1@test-20101111000000
test_target1/test_source1/fs1@test-20101111000001
test_target1/test_source1/fs1/sub
test_target1/test_source1/fs1/sub@test-20101111000000
test_target1/test_source1/fs1/sub@test-20101111000001
test_target1/test_source2
test_target1/test_source2/fs2
test_target1/test_source2/fs2/sub
test_target1/test_source2/fs2/sub@test-20101111000000
test_target1/test_source2/fs2/sub@test-20101111000001
""")


    #check consistent mounting behaviour, see issue #112
    def test_mount_consitency_mounted(self):
        """only filesystems that have canmount=on with a mountpoint should be mounted. """

        shelltest("zfs create -V 10M test_source1/fs1/subvol")

        with mocktime("20101111000000"):
            self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --allow-empty".split(" ")).run())

            r=shelltest("zfs mount |grep -o /test_target1.*")
            self.assertMultiLineEqual(r,"""
/test_target1
/test_target1/test_source1/fs1
/test_target1/test_source1/fs1/sub
/test_target1/test_source2/fs2/sub
""")


    def test_mount_consitency_unmounted(self):
        """only test_target1 should be mounted in this test"""

        shelltest("zfs create -V 10M test_source1/fs1/subvol")

        with mocktime("20101111000000"):
            self.assertFalse(ZfsAutobackup("test test_target1 --no-progress --verbose --allow-empty --clear-mountpoint".split(" ")).run())

            r=shelltest("zfs mount |grep -o /test_target1.*")
            self.assertMultiLineEqual(r,"""
/test_target1
""")




    def test_transfer_thinning(self):
        # test pre/post/during transfer thinning and efficient transfer (no transerring of stuff that gets deleted on target)

        #less output
        shelltest("zfs set autobackup:test2=true test_source1/fs1/sub")

        # nobody wants this one, will be destroyed before transferring (over a year ago)
        with mocktime("20000101000000"):
            self.assertFalse(ZfsAutobackup("test2 --allow-empty".split(" ")).run())

        # only target wants this one (monthlys)
        with mocktime("20010101000000"):
            self.assertFalse(ZfsAutobackup("test2 --allow-empty".split(" ")).run())

        # both want this one (dayly + monthly)
        # other snapshots should influence the middle one that we actually want.
        with mocktime("20010201000000"):
            shelltest("zfs snapshot test_source1/fs1/sub@other1")
            self.assertFalse(ZfsAutobackup("test2 --allow-empty".split(" ")).run())
            shelltest("zfs snapshot test_source1/fs1/sub@other2")

        # only source wants this one (dayly)
        with mocktime("20010202000000"):
            self.assertFalse(ZfsAutobackup("test2 --allow-empty".split(" ")).run())

        #will become common snapshot
        with OutputIO() as buf:
            with redirect_stdout(buf):
                with mocktime("20010203000000"):
                    self.assertFalse(ZfsAutobackup("--keep-source=1d10d --keep-target=1m10m --allow-empty --verbose --clear-mountpoint --other-snapshots test2 test_target1".split(" ")).run())


            print(buf.getvalue())
            self.assertIn(
"""
  [Source] test_source1/fs1/sub@test2-20000101000000: Destroying
  [Source] test_source1/fs1/sub@test2-20010101000000: -> test_target1/test_source1/fs1/sub (new)
  [Source] test_source1/fs1/sub@other1: -> test_target1/test_source1/fs1/sub
  [Source] test_source1/fs1/sub@test2-20010101000000: Destroying
  [Source] test_source1/fs1/sub@test2-20010201000000: -> test_target1/test_source1/fs1/sub
  [Source] test_source1/fs1/sub@other2: -> test_target1/test_source1/fs1/sub
  [Source] test_source1/fs1/sub@test2-20010203000000: -> test_target1/test_source1/fs1/sub
""", buf.getvalue())


        r=shelltest("zfs list -H -o name -r -t snapshot test_source1 test_target1")
        self.assertMultiLineEqual(r,"""
test_source1/fs1/sub@other1
test_source1/fs1/sub@test2-20010201000000
test_source1/fs1/sub@other2
test_source1/fs1/sub@test2-20010202000000
test_source1/fs1/sub@test2-20010203000000
test_target1/test_source1/fs1/sub@test2-20010101000000
test_target1/test_source1/fs1/sub@other1
test_target1/test_source1/fs1/sub@test2-20010201000000
test_target1/test_source1/fs1/sub@other2
test_target1/test_source1/fs1/sub@test2-20010203000000
""")





import itertools
import os


class TreeHasher():
    """uses BlockHasher recursively on a directory tree

    Input and output generators are in the format: ( relative-filepath, chunk_nr, hexdigest)

    """

    def __init__(self, block_hasher):
        """

        :type block_hasher: BlockHasher
        """
        self.block_hasher=block_hasher

    def generate(self, start_path):
        """Use BlockHasher on every file in a tree, yielding the results

        note that it only checks the contents of actual files. It ignores metadata like permissions and mtimes.
        It also ignores empty directories, symlinks and special files.
        """

        def walkerror(e):
            raise e

        for (dirpath, dirnames, filenames) in os.walk(start_path, onerror=walkerror):
            for f in filenames:
                file_path=os.path.join(dirpath, f)

                if (not os.path.islink(file_path)) and os.path.isfile(file_path):
                    for (chunk_nr, hash) in self.block_hasher.generate(file_path):
                        yield ( os.path.relpath(file_path,start_path), chunk_nr, hash )


    def compare(self, start_path, generator):
        """reads from generator and compares blocks

        yields mismatches in the form: ( relative_filename, chunk_nr, compare_hexdigest, actual_hexdigest )
        yields errors in the form:     ( relative_filename, chunk_nr, compare_hexdigest, "message" )

        """

        count=0

        def filter_file_name( file_name, chunk_nr, hexdigest):
                return ( chunk_nr, hexdigest )


        for file_name, group_generator in itertools.groupby(generator, lambda x: x[0]):
            count=count+1
            block_generator=itertools.starmap(filter_file_name, group_generator)
            for ( chunk_nr, compare_hexdigest, actual_hexdigest) in self.block_hasher.compare(os.path.join(start_path,file_name), block_generator):
                yield ( file_name, chunk_nr, compare_hexdigest, actual_hexdigest )







#Used for baseclasses that dont implement their own logging (Like ExecuteNode)
#Usually logging is implemented in subclasses (Like ZfsNode thats a subclass of ExecuteNode), but for regression testing its nice to have these stubs.

class LogStub:
    """Just a stub, usually overriden in subclasses."""

    # simple logging stubs
    def debug(self, txt):
        print("DEBUG  : " + txt)

    def verbose(self, txt):
        print("VERBOSE: " + txt)

    def warning(self, txt):
        print("WARNING: " + txt)

    def error(self, txt):
        print("ERROR  : " + txt)


from basetest import *
from zfs_autobackup.BlockHasher import BlockHasher


# make VERY sure this works correctly under all circumstances.

# sha1 sums of files, (bs=4096)
# da39a3ee5e6b4b0d3255bfef95601890afd80709  empty
# 642027d63bb0afd7e0ba197f2c66ad03e3d70de1  partial
# 3c0bf91170d873b8e327d3bafb6bc074580d11b7  whole
# 2e863f1fcccd6642e4e28453eba10d2d3f74d798  whole2
# 959e6b58078f0cfd2fb3d37e978fda51820473ff  whole_whole2
# 309ffffba2e1977d12f3b7469971f30d28b94bd8  whole_whole2_partial

class TestBlockHasher(unittest2.TestCase):

    def setUp(self):
        pass

    def test_empty(self):
        block_hasher = BlockHasher(count=1)
        self.assertEqual(
            list(block_hasher.generate("tests/data/empty")),
            []
        )

    def test_partial(self):
        block_hasher = BlockHasher(count=1)
        self.assertEqual(
            list(block_hasher.generate("tests/data/partial")),
            [(0, "642027d63bb0afd7e0ba197f2c66ad03e3d70de1")]
        )

    def test_whole(self):
        block_hasher = BlockHasher(count=1)
        self.assertEqual(
            list(block_hasher.generate("tests/data/whole")),
            [(0, "3c0bf91170d873b8e327d3bafb6bc074580d11b7")]
        )

    def test_whole2(self):
        block_hasher = BlockHasher(count=1)
        self.assertEqual(
            list(block_hasher.generate("tests/data/whole_whole2")),
            [
                (0, "3c0bf91170d873b8e327d3bafb6bc074580d11b7"),
                (1, "2e863f1fcccd6642e4e28453eba10d2d3f74d798")
            ]
        )

    def test_wwp(self):
        block_hasher = BlockHasher(count=1)
        self.assertEqual(
            list(block_hasher.generate("tests/data/whole_whole2_partial")),
            [
                (0, "3c0bf91170d873b8e327d3bafb6bc074580d11b7"),  # whole
                (1, "2e863f1fcccd6642e4e28453eba10d2d3f74d798"),  # whole2
                (2, "642027d63bb0afd7e0ba197f2c66ad03e3d70de1")  # partial
            ]
        )

    def test_wwp_count2(self):
        block_hasher = BlockHasher(count=2)
        self.assertEqual(
            list(block_hasher.generate("tests/data/whole_whole2_partial")),
            [
                (0, "959e6b58078f0cfd2fb3d37e978fda51820473ff"),  # whole_whole2
                (1, "642027d63bb0afd7e0ba197f2c66ad03e3d70de1")  # partial
            ]
        )

    def test_big(self):
        block_hasher = BlockHasher(count=10)
        self.assertEqual(
            list(block_hasher.generate("tests/data/whole_whole2_partial")),
            [
                (0, "309ffffba2e1977d12f3b7469971f30d28b94bd8"),  # whole_whole2_partial
            ])

    def test_blockhash_compare(self):
        #no errors
        block_hasher = BlockHasher(count=1)
        generator = block_hasher.generate("tests/data/whole_whole2_partial")
        self.assertEqual([], list(block_hasher.compare("tests/data/whole_whole2_partial", generator)))

        #compare file is smaller (EOF errors)
        block_hasher = BlockHasher(count=1)
        generator = block_hasher.generate("tests/data/whole_whole2_partial")
        self.assertEqual(
            [(1, '2e863f1fcccd6642e4e28453eba10d2d3f74d798', 'EOF'),
             (2, '642027d63bb0afd7e0ba197f2c66ad03e3d70de1', 'EOF')],
            list(block_hasher.compare("tests/data/whole", generator)))

        #no errors, huge chunks
        block_hasher = BlockHasher(count=10)
        generator = block_hasher.generate("tests/data/whole_whole2_partial")
        self.assertEqual([], list(block_hasher.compare("tests/data/whole_whole2_partial", generator)))

        # different order to make sure seek functions are ok
        block_hasher = BlockHasher(count=1)
        checksums = list(block_hasher.generate("tests/data/whole_whole2_partial"))
        checksums.reverse()
        self.assertEqual([], list(block_hasher.compare("tests/data/whole_whole2_partial", checksums)))

    def test_skip1(self):
        block_hasher = BlockHasher(count=1, skip=1)
        self.assertEqual(
            list(block_hasher.generate("tests/data/whole_whole2_partial")),
            [
                (0, "3c0bf91170d873b8e327d3bafb6bc074580d11b7"),  # whole
                # (1, "2e863f1fcccd6642e4e28453eba10d2d3f74d798"),  # whole2
                (2, "642027d63bb0afd7e0ba197f2c66ad03e3d70de1")  # partial
            ]
        )

        #should continue the pattern on the next file:
        self.assertEqual(
            list(block_hasher.generate("tests/data/whole_whole2_partial")),
            [
                # (0, "3c0bf91170d873b8e327d3bafb6bc074580d11b7"),  # whole
                (1, "2e863f1fcccd6642e4e28453eba10d2d3f74d798"),  # whole2
                # (2, "642027d63bb0afd7e0ba197f2c66ad03e3d70de1")  # partial
            ]
        )

    def test_skip6(self):
        block_hasher = BlockHasher(count=1, skip=6)
        self.assertEqual(
            list(block_hasher.generate("tests/data/whole_whole2_partial")),
            [
                (0, "3c0bf91170d873b8e327d3bafb6bc074580d11b7"),  # whole
                # (1, "2e863f1fcccd6642e4e28453eba10d2d3f74d798"),  # whole2
                # (2, "642027d63bb0afd7e0ba197f2c66ad03e3d70de1")  # partial
            ]
        )

        #all blocks of next file are skipped
        self.assertEqual(
            list(block_hasher.generate("tests/data/whole_whole2_partial")),
            [
                # (0, "3c0bf91170d873b8e327d3bafb6bc074580d11b7"),  # whole
                # (1, "2e863f1fcccd6642e4e28453eba10d2d3f74d798"),  # whole2
                # (2, "642027d63bb0afd7e0ba197f2c66ad03e3d70de1")  # partial
            ]
        )

        #first block of this one is the 6th to be skipped:
        self.assertEqual(
            list(block_hasher.generate("tests/data/whole_whole2_partial")),
            [
                # (0, "3c0bf91170d873b8e327d3bafb6bc074580d11b7"),  # whole
                (1, "2e863f1fcccd6642e4e28453eba10d2d3f74d798"),  # whole2
                # (2, "642027d63bb0afd7e0ba197f2c66ad03e3d70de1")  # partial
            ]
        )

    #NOTE: compare doesnt use skip. thats the job of its input generator


# python 2 compatibility
from __future__ import print_function
import re
import shlex
import subprocess
import sys
import time

from .ExecuteNode import ExecuteNode
from .Thinner import Thinner
from .CachedProperty import CachedProperty
from .ZfsPool import ZfsPool
from .ZfsDataset import ZfsDataset
from .ExecuteNode import ExecuteError
from .util import datetime_now


class ZfsNode(ExecuteNode):
    """a node that contains zfs datasets. implements global (systemwide/pool wide) zfs commands"""

    def __init__(self, logger, utc=False, snapshot_time_format="", hold_name="", ssh_config=None, ssh_to=None, readonly=False,
                 description="",
                 debug_output=False, thinner=None, exclude_snapshot_patterns=[]):

        self.utc = utc
        self.snapshot_time_format = snapshot_time_format
        self.hold_name = hold_name

        self.description = description

        self.logger = logger

        self.exclude_snapshot_patterns = exclude_snapshot_patterns

        if ssh_config:
            self.verbose("Using custom SSH config: {}".format(ssh_config))

        if ssh_to:
            self.verbose("SSH to: {}".format(ssh_to))
        # else:
        #     self.verbose("Datasets are local")

        if thinner is not None:
            rules = thinner.human_rules()
            if rules:
                for rule in rules:
                    self.verbose(rule)
            else:
                self.verbose("Keep no old snaphots")

        self.__thinner = thinner

        # list of ZfsPools
        self.__pools = {}
        self.__datasets = {}

        self._progress_total_bytes = 0
        self._progress_start_time = time.time()

        ExecuteNode.__init__(self, ssh_config=ssh_config, ssh_to=ssh_to, readonly=readonly, debug_output=debug_output)

    def thin(self, objects, keep_objects):
        # NOTE: if thinning is disabled with --no-thinning, self.__thinner will be none.
        if self.__thinner is not None:

            return self.__thinner.thin(objects, keep_objects, datetime_now(self.utc).timestamp())
        else:
            return (keep_objects, [])

    @CachedProperty
    def supported_send_options(self):
        """list of supported options, for optimizing sends"""
        # not every zfs implementation supports them all

        ret = []
        for option in ["-L", "-e", "-c"]:
            if self.valid_command(["zfs", "send", option, "zfs_autobackup_option_test"]):
                ret.append(option)
        return ret

    @CachedProperty
    def supported_recv_options(self):
        """list of supported options"""
        # not every zfs implementation supports them all

        ret = []
        for option in ["-s"]:
            if self.valid_command(["zfs", "recv", option, "zfs_autobackup_option_test"]):
                ret.append(option)
        return ret

    def valid_command(self, cmd):
        """test if a specified zfs options are valid exit code. use this to determine support options"""

        try:
            self.run(cmd, hide_errors=True, valid_exitcodes=[0, 1])
        except ExecuteError:
            return False

        return True

    def get_pool(self, dataset):
        """get a ZfsPool() object from dataset. stores objects internally to enable caching"""

        if not isinstance(dataset, ZfsDataset):
            raise (Exception("{} is not a ZfsDataset".format(dataset)))

        zpool_name = dataset.name.split("/")[0]

        return self.__pools.setdefault(zpool_name, ZfsPool(self, zpool_name))

    def get_dataset(self, name, force_exists=None):
        """get a ZfsDataset() object from name. stores objects internally to enable caching"""

        return self.__datasets.setdefault(name, ZfsDataset(self, name, force_exists))

    # def reset_progress(self):
    #     """reset progress output counters"""
    #     self._progress_total_bytes = 0
    #     self._progress_start_time = time.time()

    def parse_zfs_progress(self, line, hide_errors, prefix):
        """try to parse progress output of zfs recv -Pv, and don't show it as error to the user """

        # is it progress output?
        progress_fields = line.rstrip().split("\t")

        if (line.find("nvlist version") == 0 or
                line.find("resume token contents") == 0 or
                len(progress_fields) != 1 or
                line.find("skipping ") == 0 or
                re.match("send from .*estimated size is ", line)):

            # always output for debugging offcourse
            self.debug(prefix + line.rstrip())

            # actual useful info
            if len(progress_fields) >= 3:
                if progress_fields[0] == 'full' or progress_fields[0] == 'size':
                    # Reset the total bytes and start the timer again (otherwise the MB/s
                    # counter gets confused)
                    self._progress_total_bytes = int(progress_fields[2])
                    self._progress_start_time = time.time()
                elif progress_fields[0] == 'incremental':
                    # Reset the total bytes and start the timer again (otherwise the MB/s
                    # counter gets confused)
                    self._progress_total_bytes = int(progress_fields[3])
                    self._progress_start_time = time.time()
                elif progress_fields[1].isnumeric():
                    bytes_ = int(progress_fields[1])
                    if self._progress_total_bytes:
                        percentage = min(100, int(bytes_ * 100 / self._progress_total_bytes))
                        speed = int(bytes_ / (time.time() - self._progress_start_time) / (1024 * 1024))
                        bytes_left = self._progress_total_bytes - bytes_
                        minutes_left = int((bytes_left / (bytes_ / (time.time() - self._progress_start_time))) / 60)

                        self.logger.progress(
                            "Transfer {}% {}MB/s (total {}MB, {} minutes left)".format(percentage, speed, int(
                                self._progress_total_bytes / (1024 * 1024)), minutes_left))

            return

        # still do the normal stderr output handling
        if hide_errors:
            self.debug(prefix + line.rstrip())
        else:
            self.error(prefix + line.rstrip())

    # def _parse_stderr_pipe(self, line, hide_errors):
    #     self.parse_zfs_progress(line, hide_errors, "STDERR|> ")

    def _parse_stderr(self, line, hide_errors):
        self.parse_zfs_progress(line, hide_errors, "STDERR > ")

    def verbose(self, txt):
        self.logger.verbose("{} {}".format(self.description, txt))

    def error(self, txt):
        self.logger.error("{} {}".format(self.description, txt))

    def warning(self, txt):
        self.logger.warning("{} {}".format(self.description, txt))

    def debug(self, txt):
        self.logger.debug("{} {}".format(self.description, txt))

    def consistent_snapshot(self, datasets, snapshot_name, min_changed_bytes, pre_snapshot_cmds=[],
                            post_snapshot_cmds=[], set_snapshot_properties=[]):
        """create a consistent (atomic) snapshot of specified datasets, per pool.
        """

        pools = {}

        # collect snapshots that we want to make, per pool
        # self.debug(datasets)
        for dataset in datasets:
            if not dataset.is_changed_ours(min_changed_bytes):
                dataset.verbose("No changes since {}".format(dataset.our_snapshots[-1].snapshot_name))
                continue

            # force_exist, since we're making it
            snapshot = self.get_dataset(dataset.name + "@" + snapshot_name, force_exists=True)

            pool = dataset.split_path()[0]
            if pool not in pools:
                pools[pool] = []

            pools[pool].append(snapshot)

            # update cache, but try to prevent an unneeded zfs list
            if self.readonly or CachedProperty.is_cached(dataset, 'snapshots'):
                dataset.snapshots.append(snapshot)  # NOTE: this will trigger zfs list if its not cached

        if not pools:
            self.verbose("No changes anywhere: not creating snapshots.")
            return

        try:
            for cmd in pre_snapshot_cmds:
                self.verbose("Running pre-snapshot-cmd")
                self.run(cmd=shlex.split(cmd), readonly=False)

            # create consistent snapshot per pool
            for (pool_name, snapshots) in pools.items():
                cmd = ["zfs", "snapshot"]
                for snapshot_property in set_snapshot_properties:
                    cmd += ['-o', snapshot_property]

                cmd.extend(map(lambda snapshot_: str(snapshot_), snapshots))

                self.verbose("Creating snapshots {} in pool {}".format(snapshot_name, pool_name))
                self.run(cmd, readonly=False)

        finally:
            for cmd in post_snapshot_cmds:
                self.verbose("Running post-snapshot-cmd")
                try:
                    self.run(cmd=shlex.split(cmd), readonly=False)
                except Exception as e:
                    pass

    def selected_datasets(self, property_name, exclude_received, exclude_paths, exclude_unchanged):
        """determine filesystems that should be backed up by looking at the special autobackup-property, systemwide

           returns: ( list of selected ZfsDataset, list of excluded ZfsDataset)
        """

        self.debug("Getting selected datasets")

        # get all source filesystems that have the backup property
        lines = self.run(tab_split=True, readonly=True, cmd=[
            "zfs", "get", "-t", "volume,filesystem", "-o", "name,value,source", "-H",
            property_name
        ])


        # The returnlist of selected ZfsDataset's:
        selected_filesystems = []
        excluded_filesystems = []

        # list of sources, used to resolve inherited sources
        sources = {}

        for line in lines:
            (name, value, raw_source) = line
            dataset = self.get_dataset(name, force_exists=True)

            # "resolve" inherited sources
            sources[name] = raw_source
            if raw_source.find("inherited from ") == 0:
                inherited = True
                inherited_from = re.sub("^inherited from ", "", raw_source)
                source = sources[inherited_from]
            else:
                inherited = False
                source = raw_source

            # determine it
            selected=dataset.is_selected(value=value, source=source, inherited=inherited, exclude_received=exclude_received,
                                   exclude_paths=exclude_paths, exclude_unchanged=exclude_unchanged)

            if selected==True:
                selected_filesystems.append(dataset)
            elif selected==False:
                excluded_filesystems.append(dataset)
            #returns None when no property is set.


        return ( selected_filesystems, excluded_filesystems)




from basetest import *


# test zfs-verify:
# - when there is no common snapshot at all
# - when encryption key not loaded
# - --test mode
# - --fs-compare methods
# - on snapshots of datasets:
#   - that are correct
#   - that are different
# - on snapshots of zvols
#  - that are correct
#  - that are different
# - test all directions (local, remote/local, local/remote, remote/remote)
#

class TestZfsVerify(unittest2.TestCase):


    def setUp(self):
        self.skipTest("WIP")

        prepare_zpools()

        #create actual test files and data
        shelltest("zfs create test_source1/fs1/ok_filesystem")
        shelltest("cp tests/*.py /test_source1/fs1/ok_filesystem")

        shelltest("zfs create test_source1/fs1/bad_filesystem")
        shelltest("cp tests/*.py /test_source1/fs1/bad_filesystem")

        shelltest("zfs create -V 1M test_source1/fs1/ok_zvol")
        shelltest("dd if=/dev/urandom of=/dev/zvol/test_source1/fs1/ok_zvol count=1 bs=512k")

        shelltest("zfs create -V 1M test_source1/fs1/bad_zvol")
        shelltest("dd if=/dev/urandom of=/dev/zvol/test_source1/fs1/bad_zvol count=1 bs=512k")

        #create backup
        with mocktime("20101111000000"):
            self.assertFalse(ZfsAutobackup("test test_target1 --verbose --no-progress --no-holds".split(" ")).run())

        #Do an ugly hack to create a fault in the bad filesystem
        #In zfs-autoverify it doenst matter that the snapshot isnt actually the same snapshot, so this hack works
        shelltest("zfs destroy test_target1/test_source1/fs1/bad_filesystem@test-20101111000000")
        shelltest("zfs mount test_target1/test_source1/fs1/bad_filesystem")
        shelltest("echo >> /test_target1/test_source1/fs1/bad_filesystem/test_verify.py")
        shelltest("zfs snapshot test_target1/test_source1/fs1/bad_filesystem@test-20101111000000")

        #do the same hack for the bad zvol
        shelltest("zfs destroy test_target1/test_source1/fs1/bad_zvol@test-20101111000000")
        shelltest("dd if=/dev/urandom of=/dev/zvol/test_target1/test_source1/fs1/bad_zvol count=1 bs=1")
        shelltest("zfs snapshot test_target1/test_source1/fs1/bad_zvol@test-20101111000000")


        # make sure we cant accidently compare current data
        shelltest("zfs mount test_target1/test_source1/fs1/ok_filesystem")
        shelltest("rm /test_source1/fs1/ok_filesystem/*")
        shelltest("rm /test_source1/fs1/bad_filesystem/*")
        shelltest("dd if=/dev/zero of=/dev/zvol/test_source1/fs1/ok_zvol count=1 bs=512k")



    def test_verify(self):


        with self.subTest("default --test"):
            self.assertFalse(ZfsAutoverify("test test_target1 --verbose --test".split(" ")).run())

        with self.subTest("rsync, remote source and target. (not supported, all 6 fail)"):
            self.assertEqual(6, ZfsAutoverify("test test_target1 --ssh-source=localhost --ssh-target=localhost --verbose --exclude-received --fs-compare=rsync".split(" ")).run())

        def runchecked(testname, command):
            with self.subTest(testname):
                with OutputIO() as buf:
                    result=None
                    with redirect_stderr(buf):
                        result=ZfsAutoverify(command.split(" ")).run()

                    print(buf.getvalue())
                    self.assertEqual(2,result)
                    self.assertRegex(buf.getvalue(), "bad_filesystem: FAILED:")
                    self.assertRegex(buf.getvalue(), "bad_zvol: FAILED:")

        runchecked("rsync, remote source", "test test_target1 --ssh-source=localhost --verbose --exclude-received --fs-compare=rsync")
        runchecked("rsync, remote target", "test test_target1 --ssh-target=localhost --verbose --exclude-received --fs-compare=rsync")
        runchecked("rsync, local", "test test_target1 --verbose --exclude-received --fs-compare=rsync")

        runchecked("tar, remote source and remote target",
                   "test test_target1 --ssh-source=localhost --ssh-target=localhost --verbose --exclude-received --fs-compare=find")
        runchecked("tar, remote source",
                   "test test_target1 --ssh-source=localhost --verbose --exclude-received --fs-compare=find")
        runchecked("tar, remote target",
                   "test test_target1 --ssh-target=localhost --verbose --exclude-received --fs-compare=find")
        runchecked("tar, local", "test test_target1 --verbose --exclude-received --fs-compare=find")

        with self.subTest("no common snapshot"):
            #destroy common snapshot, now 3 should fail
            shelltest("zfs destroy test_source1/fs1/ok_zvol@test-20101111000000")
            self.assertEqual(3, ZfsAutoverify("test test_target1 --verbose --exclude-received".split(" ")).run())




import argparse
import re
import sys

from .CliBase import CliBase
from .util import datetime_now


class ZfsAuto(CliBase):
    """Common Base class for ZfsAutobackup and ZfsAutoverify ."""

    def __init__(self, argv, print_arguments=True):

        self.hold_name = None
        self.snapshot_time_format = None
        self.property_name = None
        self.exclude_paths = None

        super(ZfsAuto, self).__init__(argv, print_arguments)

    def parse_args(self, argv):
        """parse common arguments, setup logging, check and adjust parameters"""

        args = super(ZfsAuto, self).parse_args(argv)

        if args.backup_name == None:
            self.parser.print_usage()
            self.log.error("Please specify BACKUP-NAME")
            sys.exit(255)

        if args.target_path is not None and args.target_path[0] == "/":
            self.log.error("Target should not start with a /")
            sys.exit(255)

        if args.ignore_replicated:
            self.warning("--ignore-replicated has been renamed, using --exclude-unchanged")
            args.exclude_unchanged = True

        # Note: Before version v3.1-beta5, we always used exclude_received. This was a problem if you wanted to
        # replicate an existing backup to another host and use the same backupname/snapshots. However, exclude_received
        # may still need to be used to explicitly exclude a backup with the 'received' source property to avoid accidental
        # recursive replication of a zvol that is currently being received in another session (as it will have changes).

        self.exclude_paths = []
        if args.ssh_source == args.ssh_target:
            if args.target_path:
                # target and source are the same, make sure to exclude target_path
                self.verbose("NOTE: Source and target are on the same host, excluding target-path from selection.")
                self.exclude_paths.append(args.target_path)
            else:
                if not args.exclude_received and not args.include_received:
                    self.verbose("NOTE: Source and target are on the same host, adding --exclude-received to commandline. (use --include-received to overrule)")
                    args.exclude_received = True

        if args.test:
            self.warning("TEST MODE - SIMULATING WITHOUT MAKING ANY CHANGES")

        #format all the names
        self.property_name = args.property_format.format(args.backup_name)
        self.snapshot_time_format = args.snapshot_format.format(args.backup_name)
        self.hold_name = args.hold_format.format(args.backup_name)

        dt = datetime_now(args.utc)

        self.verbose("")
        self.verbose("Current time {}           : {}".format(args.utc and "UTC" or "   ", dt.strftime("%Y-%m-%d %H:%M:%S")))

        self.verbose("Selecting dataset property : {}".format(self.property_name))
        self.verbose("Snapshot format            : {}".format(self.snapshot_time_format))
        self.verbose("Timezone                   : {}".format("UTC" if args.utc else "Local"))

        return args

    def get_parser(self):

        parser = super(ZfsAuto, self).get_parser()

        #positional arguments
        parser.add_argument('backup_name', metavar='BACKUP-NAME', default=None, nargs='?',
                            help='Name of the backup to select')

        parser.add_argument('target_path', metavar='TARGET-PATH', default=None, nargs='?',
                            help='Target ZFS filesystem (optional)')



        # SSH options
        group=parser.add_argument_group("SSH options")
        group.add_argument('--ssh-config', metavar='CONFIG-FILE', default=None, help='Custom ssh client config')
        group.add_argument('--ssh-source', metavar='USER@HOST', default=None,
                            help='Source host to pull backup from.')
        group.add_argument('--ssh-target', metavar='USER@HOST', default=None,
                            help='Target host to push backup to.')

        group=parser.add_argument_group("String formatting options")
        group.add_argument('--property-format', metavar='FORMAT', default="autobackup:{}",
                            help='Dataset selection string format. Default: %(default)s')
        group.add_argument('--snapshot-format', metavar='FORMAT', default="{}-%Y%m%d%H%M%S",
                            help='ZFS Snapshot string format. Default: %(default)s')
        group.add_argument('--hold-format', metavar='FORMAT', default="zfs_autobackup:{}",
                            help='ZFS hold string format. Default: %(default)s')
        group.add_argument('--strip-path', metavar='N', default=0, type=int,
                           help='Number of directories to strip from target path.')

        group=parser.add_argument_group("Selection options")
        group.add_argument('--ignore-replicated', action='store_true', help=argparse.SUPPRESS)
        group.add_argument('--exclude-unchanged', metavar='BYTES', default=0, type=int,
                            help='Exclude datasets that have less than BYTES data changed since any last snapshot. (Use with proxmox HA replication)')
        group.add_argument('--exclude-received', action='store_true',
                            help='Exclude datasets that have the origin of their autobackup: property as "received". '
                                 'This can avoid recursive replication between two backup partners.')
        group.add_argument('--include-received', action='store_true',
                            help=argparse.SUPPRESS)


        def regex_argument_type(input_line):
            """Parses regex arguments into re.Pattern objects"""
            try:
                return re.compile(input_line)
            except:
                raise ValueError("Could not parse argument '{}' as a regular expression".format(input_line))
        group.add_argument('--exclude-snapshot-pattern', action='append', default=[], type=regex_argument_type, help="Regular expression to match snapshots that will be ignored.")

        return parser

    def print_error_sources(self):
        self.error(
            "No source filesystems selected, please do a 'zfs set autobackup:{0}=true' on the source datasets "
            "you want to select.".format(
                self.args.backup_name))

    def make_target_name(self, source_dataset):
        """make target_name from a source_dataset"""
        return self.args.target_path + "/" + source_dataset.lstrip_path(self.args.strip_path)



from basetest import *
from zfs_autobackup.LogStub import LogStub
from zfs_autobackup.ExecuteNode import ExecuteError


class TestZfsNode(unittest2.TestCase):

    def setUp(self):
        prepare_zpools()
        # return super().setUp()

    def test_consistent_snapshot(self):
        logger = LogStub()
        description = "[Source]"
        node = ZfsNode(utc=False, snapshot_time_format="test-%Y%m%d%H%M%S", hold_name="zfs_autobackup:test", logger=logger, description=description)

        with self.subTest("first snapshot"):
            (selected_datasets, excluded_datasets)=node.selected_datasets(property_name="autobackup:test", exclude_paths=[], exclude_received=False,
                                   exclude_unchanged=0)
            node.consistent_snapshot(selected_datasets, "test-20101111000001", 100000)
            r = shelltest("zfs list -H -o name -r -t all " + TEST_POOLS)
            self.assertEqual(r, """
test_source1
test_source1/fs1
test_source1/fs1@test-20101111000001
test_source1/fs1/sub
test_source1/fs1/sub@test-20101111000001
test_source2
test_source2/fs2
test_source2/fs2/sub
test_source2/fs2/sub@test-20101111000001
test_source2/fs3
test_source2/fs3/sub
test_target1
""")

        with self.subTest("second snapshot, no changes, no snapshot"):
            (selected_datasets, excluded_datasets)=node.selected_datasets(property_name="autobackup:test", exclude_paths=[], exclude_received=False,
                                   exclude_unchanged=0)
            node.consistent_snapshot(selected_datasets, "test-20101111000002", 1)
            r = shelltest("zfs list -H -o name -r -t all " + TEST_POOLS)
            self.assertEqual(r, """
test_source1
test_source1/fs1
test_source1/fs1@test-20101111000001
test_source1/fs1/sub
test_source1/fs1/sub@test-20101111000001
test_source2
test_source2/fs2
test_source2/fs2/sub
test_source2/fs2/sub@test-20101111000001
test_source2/fs3
test_source2/fs3/sub
test_target1
""")

        with self.subTest("second snapshot, no changes, empty snapshot"):
            (selected_datasets, excluded_datasets) =node.selected_datasets(property_name="autobackup:test", exclude_paths=[], exclude_received=False, exclude_unchanged=0)
            node.consistent_snapshot(selected_datasets, "test-20101111000002", 0)
            r = shelltest("zfs list -H -o name -r -t all " + TEST_POOLS)
            self.assertEqual(r, """
test_source1
test_source1/fs1
test_source1/fs1@test-20101111000001
test_source1/fs1@test-20101111000002
test_source1/fs1/sub
test_source1/fs1/sub@test-20101111000001
test_source1/fs1/sub@test-20101111000002
test_source2
test_source2/fs2
test_source2/fs2/sub
test_source2/fs2/sub@test-20101111000001
test_source2/fs2/sub@test-20101111000002
test_source2/fs3
test_source2/fs3/sub
test_target1
""")

    def test_consistent_snapshot_prepostcmds(self):
        logger = LogStub()
        description = "[Source]"
        node = ZfsNode(utc=False, snapshot_time_format="test", hold_name="test", logger=logger, description=description, debug_output=True)

        with self.subTest("Test if all cmds are executed correctly (no failures)"):
            with OutputIO() as buf:
                with redirect_stdout(buf):
                    (selected_datasets, excluded_datasets) =node.selected_datasets(property_name="autobackup:test", exclude_paths=[], exclude_received=False, exclude_unchanged=0)
                    node.consistent_snapshot(selected_datasets, "test-1",
                                             0,
                                             pre_snapshot_cmds=["echo pre1", "echo pre2"],
                                             post_snapshot_cmds=["echo post1 >&2", "echo post2 >&2"]
                                             )

                self.assertIn("STDOUT > pre1", buf.getvalue())
                self.assertIn("STDOUT > pre2", buf.getvalue())
                self.assertIn("STDOUT > post1", buf.getvalue())
                self.assertIn("STDOUT > post2", buf.getvalue())


        with self.subTest("Failure in the middle, only pre1 and both post1 and post2 should be executed, no snapshot should be attempted"):
            with OutputIO() as buf:
                with redirect_stdout(buf):
                    with self.assertRaises(ExecuteError):
                        (selected_datasets, excluded_datasets) =node.selected_datasets(property_name="autobackup:test", exclude_paths=[], exclude_received=False, exclude_unchanged=0)
                        node.consistent_snapshot(selected_datasets, "test-1",
                                                 0,
                                                 pre_snapshot_cmds=["echo pre1", "false", "echo pre2"],
                                                 post_snapshot_cmds=["echo post1", "false", "echo post2"]
                                                 )

                print(buf.getvalue())
                self.assertIn("STDOUT > pre1", buf.getvalue())
                self.assertNotIn("STDOUT > pre2", buf.getvalue())
                self.assertIn("STDOUT > post1", buf.getvalue())
                self.assertIn("STDOUT > post2", buf.getvalue())

        with self.subTest("Snapshot fails"):
            with OutputIO() as buf:
                with redirect_stdout(buf):
                    with self.assertRaises(ExecuteError):
                        #same snapshot name as before so it fails
                        (selected_datasets, excluded_datasets) =node.selected_datasets(property_name="autobackup:test", exclude_paths=[], exclude_received=False, exclude_unchanged=0)
                        node.consistent_snapshot(selected_datasets, "test-1",
                                                 0,
                                                 pre_snapshot_cmds=["echo pre1", "echo pre2"],
                                                 post_snapshot_cmds=["echo post1", "echo post2"]
                                                 )

                print(buf.getvalue())
                self.assertIn("STDOUT > pre1", buf.getvalue())
                self.assertIn("STDOUT > pre2", buf.getvalue())
                self.assertIn("STDOUT > post1", buf.getvalue())
                self.assertIn("STDOUT > post2", buf.getvalue())

    def test_timestamps(self):
        # Assert that timestamps keep relative order both for utc and for localtime
        logger = LogStub()
        description = "[Source]"
        node_local = ZfsNode(utc=False, snapshot_time_format="test-%Y%m%d%H%M%S", hold_name="zfs_autobackup:test", logger=logger, description=description)
        node_utc = ZfsNode(utc=True, snapshot_time_format="test-%Y%m%d%H%M%S", hold_name="zfs_autobackup:test", logger=logger, description=description)

        for node in [node_local, node_utc]:
            with self.subTest("timestamp ordering " + ("utc" if node == node_utc else "localtime")):
                dataset_a = ZfsDataset(node,"test_source1@test-20101111000001")
                dataset_b = ZfsDataset(node,"test_source1@test-20101111000002")
                dataset_c = ZfsDataset(node,"test_source1@test-20240101020202")
                self.assertGreater(dataset_b.timestamp, dataset_a.timestamp)
                self.assertGreater(dataset_c.timestamp, dataset_b.timestamp)


    def test_getselected(self):

        # should be excluded by property
        shelltest("zfs create test_source1/fs1/subexcluded")
        shelltest("zfs set autobackup:test=false test_source1/fs1/subexcluded")

        # only select parent
        shelltest("zfs create test_source1/fs1/onlyparent")
        shelltest("zfs create test_source1/fs1/onlyparent/child")
        shelltest("zfs set autobackup:test=parent test_source1/fs1/onlyparent")

        # should be excluded by being unchanged
        shelltest("zfs create test_source1/fs1/unchanged")
        shelltest("zfs snapshot test_source1/fs1/unchanged@somesnapshot")

        logger = LogStub()
        description = "[Source]"
        node = ZfsNode(utc=False, snapshot_time_format="test-%Y%m%d%H%M%S", hold_name="zfs_autobackup:test", logger=logger, description=description)
        (selected_datasets, excluded_datasets)=node.selected_datasets(property_name="autobackup:test", exclude_paths=[], exclude_received=False,
                               exclude_unchanged=1)
        s = pformat(selected_datasets)
        print(s)

        # basics
        self.assertEqual(s, """[(local): test_source1/fs1,
 (local): test_source1/fs1/onlyparent,
 (local): test_source1/fs1/sub,
 (local): test_source2/fs2/sub]""")


    def test_validcommand(self):
        logger = LogStub()
        description = "[Source]"
        node = ZfsNode(utc=False, snapshot_time_format="test-%Y%m%d%H%M%S", hold_name="zfs_autobackup:test", logger=logger, description=description)

        with self.subTest("test invalid option"):
            self.assertFalse(node.valid_command(["zfs", "send", "--invalid-option", "nonexisting"]))
        with self.subTest("test valid option"):
            self.assertTrue(node.valid_command(["zfs", "send", "-v", "nonexisting"]))

    def test_supportedsendoptions(self):
        logger = LogStub()
        description = "[Source]"
        node = ZfsNode(utc=False, snapshot_time_format="test-%Y%m%d%H%M%S", hold_name="zfs_autobackup:test", logger=logger, description=description)
        # -D propably always supported
        self.assertGreater(len(node.supported_send_options), 0)

    def test_supportedrecvoptions(self):
        logger = LogStub()
        description = "[Source]"
        # NOTE: this could hang via ssh if we dont close filehandles properly. (which was a previous bug)
        node = ZfsNode(utc=False, snapshot_time_format="test-%Y%m%d%H%M%S", hold_name="zfs_autobackup:test", logger=logger, description=description, ssh_to='localhost')
        self.assertIsInstance(node.supported_recv_options, list)


if __name__ == '__main__':
    unittest.main()




import argparse
from signal import signal, SIGPIPE
from .util import output_redir, sigpipe_handler, datetime_now

from .ZfsAuto import ZfsAuto

from . import compressors
from .ExecuteNode import ExecuteNode
from .Thinner import Thinner
from .ZfsDataset import ZfsDataset
from .ZfsNode import ZfsNode
from .ThinnerRule import ThinnerRule

class ZfsAutobackup(ZfsAuto):
    """The main zfs-autobackup class. Start here, at run() :)"""

    def __init__(self, argv, print_arguments=True):

        # NOTE: common options and parameters are in ZfsAuto
        super(ZfsAutobackup, self).__init__(argv, print_arguments)

    def parse_args(self, argv):
        """do extra checks on common args"""

        args = super(ZfsAutobackup, self).parse_args(argv)

        if not args.no_holds:
            self.verbose("Hold name                  : {}".format(self.hold_name))

        if args.allow_empty:
            args.min_change = 0

        # if args.destroy_incompatible:
        #     args.rollback = True

        if args.resume:
            self.warning("The --resume option isn't needed anymore (it's autodetected now)")

        if args.raw:
            self.warning(
                "The --raw option isn't needed anymore (it's autodetected now). Also see --encrypt and --decrypt.")

        if args.compress and args.ssh_source is None and args.ssh_target is None:
            self.warning("Using compression, but transfer is local.")

        if args.compress and args.zfs_compressed:
            self.warning("Using --compress with --zfs-compressed, might be inefficient.")

        return args

    def get_parser(self):
        """extend common parser with  extra stuff needed for zfs-autobackup"""

        parser = super(ZfsAutobackup, self).get_parser()

        group = parser.add_argument_group("Snapshot options")
        group.add_argument('--no-snapshot', action='store_true',
                           help='Don\'t create new snapshots (useful for finishing uncompleted backups, or cleanups)')
        group.add_argument('--pre-snapshot-cmd', metavar="COMMAND", default=[], action='append',
                           help='Run COMMAND before snapshotting (can be used multiple times.')
        group.add_argument('--post-snapshot-cmd', metavar="COMMAND", default=[], action='append',
                           help='Run COMMAND after snapshotting (can be used multiple times.')
        group.add_argument('--min-change', metavar='BYTES', type=int, default=1,
                           help='Only create snapshot if enough bytes are changed. (default %('
                                'default)s)')
        group.add_argument('--allow-empty', action='store_true',
                           help='If nothing has changed, still create empty snapshots. (Same as --min-change=0)')
        group.add_argument('--other-snapshots', action='store_true',
                           help='Send over other snapshots as well, not just the ones created by this tool.')
        group.add_argument('--set-snapshot-properties', metavar='PROPERTY=VALUE,...', type=str,
                           help='List of properties to set on the snapshot.')
        group.add_argument('--no-guid-check', action='store_true',
                           help='Dont check guid of common snapshots. (faster)')


        group = parser.add_argument_group("Transfer options")
        group.add_argument('--no-send', action='store_true',
                           help='Don\'t transfer snapshots (useful for cleanups, or if you want a separate send-cronjob)')
        group.add_argument('--no-holds', action='store_true',
                           help='Don\'t hold snapshots. (Faster. Allows you to destroy common snapshot.)')
        group.add_argument('--clear-refreservation', action='store_true',
                           help='Filter "refreservation" property. (recommended, saves space. same as '
                                '--filter-properties refreservation)')
        group.add_argument('--clear-mountpoint', action='store_true',
                           help='Set property canmount=noauto for new datasets. (recommended, prevents mount '
                                'conflicts. same as --set-properties canmount=noauto)')
        group.add_argument('--filter-properties', metavar='PROPERTY,...', type=str,
                           help='List of properties to "filter" when receiving filesystems. (you can still restore '
                                'them with zfs inherit -S)')
        group.add_argument('--set-properties', metavar='PROPERTY=VALUE,...', type=str,
                           help='List of propererties to override when receiving filesystems. (you can still restore '
                                'them with zfs inherit -S)')
        group.add_argument('--rollback', action='store_true',
                           help='Rollback changes to the latest target snapshot before starting. (normally you can '
                                'prevent changes by setting the readonly property on the target_path to on)')
        group.add_argument('--force', '-F', action='store_true',
                           help='Use zfs -F option to force overwrite/rollback. (Useful with --strip-path=1, but use with care)')
        group.add_argument('--destroy-incompatible', action='store_true',
                           help='Destroy incompatible snapshots on target. Use with care! (also does rollback of dataset)')
        group.add_argument('--ignore-transfer-errors', action='store_true',
                           help='Ignore transfer errors (still checks if received filesystem exists. useful for '
                                'acltype errors)')

        group.add_argument('--decrypt', action='store_true',
                           help='Decrypt data before sending it over.')
        group.add_argument('--encrypt', action='store_true',
                           help='Encrypt data after receiving it.')

        group.add_argument('--zfs-compressed', action='store_true',
                           help='Transfer blocks that already have zfs-compression as-is.')

        group = parser.add_argument_group("Data transfer options")
        group.add_argument('--compress', metavar='TYPE', default=None, nargs='?', const='zstd-fast',
                           choices=compressors.choices(),
                           help='Use compression during transfer, defaults to zstd-fast if TYPE is not specified. ({})'.format(
                               ", ".join(compressors.choices())))
        group.add_argument('--rate', metavar='DATARATE', default=None,
                           help='Limit data transfer rate in Bytes/sec (e.g. 128K. requires mbuffer.)')
        group.add_argument('--buffer', metavar='SIZE', default=None,
                           help='Add zfs send and recv buffers to smooth out IO bursts. (e.g. 128M. requires mbuffer)')
        parser.add_argument('--buffer-chunk-size', metavar="BUFFERCHUNKSIZE", default=None,
                            help='Tune chunk size when mbuffer is used. (requires mbuffer.)')
        group.add_argument('--send-pipe', metavar="COMMAND", default=[], action='append',
                           help='pipe zfs send output through COMMAND (can be used multiple times)')
        group.add_argument('--recv-pipe', metavar="COMMAND", default=[], action='append',
                           help='pipe zfs recv input through COMMAND (can be used multiple times)')

        group = parser.add_argument_group("Thinner options")
        group.add_argument('--no-thinning', action='store_true', help="Do not destroy any snapshots.")
        group.add_argument('--keep-source', metavar='SCHEDULE', type=str, default="10,1d1w,1w1m,1m1y",
                           help='Thinning schedule for old source snapshots. Default: %(default)s')
        group.add_argument('--keep-target', metavar='SCHEDULE', type=str, default="10,1d1w,1w1m,1m1y",
                           help='Thinning schedule for old target snapshots. Default: %(default)s')
        group.add_argument('--destroy-missing', metavar="SCHEDULE", type=str, default=None,
                           help='Destroy datasets on target that are missing on the source. Specify the time since '
                                'the last snapshot, e.g: --destroy-missing 30d')

        # obsolete
        parser.add_argument('--resume', action='store_true', help=argparse.SUPPRESS)
        parser.add_argument('--raw', action='store_true', help=argparse.SUPPRESS)

        return parser

    # NOTE: this method also uses self.args. args that need extra processing are passed as function parameters:
    def thin_missing_targets(self, target_dataset, used_target_datasets):
        """thin target datasets that are missing on the source.
        :type used_target_datasets: list[ZfsDataset]
        :type target_dataset: ZfsDataset
        """

        self.debug("Thinning obsolete datasets")
        missing_datasets = [dataset for dataset in target_dataset.recursive_datasets if
                            dataset not in used_target_datasets]

        count = 0
        for dataset in missing_datasets:
            self.debug("analyse missing {}".format(dataset))

            count = count + 1
            if self.args.progress:
                self.progress("Analysing missing {}/{}".format(count, len(missing_datasets)))

            try:
                dataset.debug("Missing on source, thinning")
                dataset.thin()

            except Exception as e:
                dataset.error("Error during thinning of missing datasets ({})".format(str(e)))

        # if self.args.progress:
        #     self.clear_progress()

    # NOTE: this method also uses self.args. args that need extra processing are passed as function parameters:
    def destroy_missing_targets(self, target_dataset, used_target_datasets):
        """destroy target datasets that are missing on the source and that meet the requirements
        :type used_target_datasets: list[ZfsDataset]
        :type target_dataset: ZfsDataset

        """

        self.debug("Destroying obsolete datasets")

        missing_datasets = [dataset for dataset in target_dataset.recursive_datasets if
                            dataset not in used_target_datasets]

        count = 0
        for dataset in missing_datasets:

            count = count + 1
            if self.args.progress:
                self.progress("Analysing destroy missing {}/{}".format(count, len(missing_datasets)))

            try:
                # cant do anything without our own snapshots
                if not dataset.our_snapshots:
                    if dataset.datasets:
                        # its not a leaf, just ignore
                        dataset.debug("Destroy missing: ignoring")
                    else:
                        dataset.verbose(
                            "Destroy missing: has no snapshots made by us (please destroy manually).")
                else:
                    # past the deadline?
                    deadline_ttl = ThinnerRule("0s" + self.args.destroy_missing).ttl
                    now = datetime_now(self.args.utc).timestamp()
                    if dataset.our_snapshots[-1].timestamp + deadline_ttl > now:
                        dataset.verbose("Destroy missing: Waiting for deadline.")
                    else:

                        dataset.debug("Destroy missing: Removing our snapshots.")

                        # remove all our snaphots, except last, to safe space in case we fail later on
                        for snapshot in dataset.our_snapshots[:-1]:
                            snapshot.destroy(fail_exception=True)

                        # does it have other snapshots?
                        has_others = False
                        for snapshot in dataset.snapshots:
                            if not snapshot.is_ours():
                                has_others = True
                                break

                        if has_others:
                            dataset.verbose("Destroy missing: Still in use by other snapshots")
                        else:
                            if dataset.datasets:
                                dataset.verbose("Destroy missing: Still has children here.")
                            else:
                                dataset.verbose("Destroy missing.")
                                dataset.our_snapshots[-1].destroy(fail_exception=True)
                                dataset.destroy(fail_exception=True)

            except Exception as e:
                # if self.args.progress:
                #     self.clear_progress()

                dataset.error("Error during --destroy-missing: {}".format(str(e)))

        # if self.args.progress:
        #     self.clear_progress()

    def get_send_pipes(self, logger):
        """determine the zfs send pipe"""

        ret = []
        _mbuffer = False
        _buffer = "16M"
        _cs = "128k"
        _rate = False

        # IO buffer
        if self.args.buffer:
            logger("zfs send buffer        : {}".format(self.args.buffer))
            _mbuffer = True
            _buffer = self.args.buffer

        # IO chunk size
        if self.args.buffer_chunk_size:
            logger("zfs send chunk size    : {}".format(self.args.buffer_chunk_size))
            _mbuffer = True
            _cs = self.args.buffer_chunk_size

        # custom pipes
        for send_pipe in self.args.send_pipe:
            ret.append(ExecuteNode.PIPE)
            ret.extend(send_pipe.split(" "))
            logger("zfs send custom pipe   : {}".format(send_pipe))

        # compression
        if self.args.compress != None:
            ret.append(ExecuteNode.PIPE)
            cmd = compressors.compress_cmd(self.args.compress)
            ret.extend(cmd)
            logger("zfs send compression   : {}".format(" ".join(cmd)))

        # transfer rate
        if self.args.rate:
            logger("zfs send transfer rate : {}".format(self.args.rate))
            _mbuffer = True
            _rate = self.args.rate

        if _mbuffer:
            cmd = [ExecuteNode.PIPE, "mbuffer", "-q", "-s{}".format(_cs), "-m{}".format(_buffer)]
            if _rate:
                cmd.append("-R{}".format(self.args.rate))
            ret.extend(cmd)

        return ret

    def get_recv_pipes(self, logger):

        ret = []

        # decompression
        if self.args.compress != None:
            cmd = compressors.decompress_cmd(self.args.compress)
            ret.extend(cmd)
            ret.append(ExecuteNode.PIPE)
            logger("zfs recv decompression : {}".format(" ".join(cmd)))

        # custom pipes
        for recv_pipe in self.args.recv_pipe:
            ret.extend(recv_pipe.split(" "))
            ret.append(ExecuteNode.PIPE)
            logger("zfs recv custom pipe   : {}".format(recv_pipe))

        # IO buffer
        if self.args.buffer or self.args.buffer_chunk_size:
            _cs = "128k"
            _buffer = "16M"
            # only add second buffer if its usefull. (e.g. non local transfer or other pipes active)
            if self.args.ssh_source != None or self.args.ssh_target != None or self.args.recv_pipe or self.args.send_pipe or self.args.compress != None:
                logger("zfs recv buffer        : {}".format(self.args.buffer))

                if self.args.buffer_chunk_size:
                    _cs = self.args.buffer_chunk_size
                if self.args.buffer:
                    _buffer = self.args.buffer

                ret.extend(["mbuffer", "-q", "-s{}".format(_cs), "-m{}".format(_buffer), ExecuteNode.PIPE])

        return ret

    def make_target_name(self, source_dataset):
        """make target_name from a source_dataset"""
        stripped=source_dataset.lstrip_path(self.args.strip_path)
        if stripped!="":
            return self.args.target_path + "/" + stripped
        else:
            return self.args.target_path

    def check_target_names(self, source_node, source_datasets, target_node):
        """check all target names for collesions etc due to strip-options"""

        self.debug("Checking target names:")
        target_datasets={}
        for source_dataset in source_datasets:

            target_name = self.make_target_name(source_dataset)
            source_dataset.debug("-> {}".format(target_name))

            if target_name in target_datasets:
                raise Exception("Target collision: Target path {} encountered twice, due to: {} and {}".format(target_name, source_dataset, target_datasets[target_name]))

            target_datasets[target_name]=source_dataset

    # NOTE: this method also uses self.args. args that need extra processing are passed as function parameters:
    def sync_datasets(self, source_node, source_datasets, target_node):
        """Sync datasets, or thin-only on both sides
        :type target_node: ZfsNode
        :type source_datasets: list of ZfsDataset
        :type source_node: ZfsNode
        """

        send_pipes = self.get_send_pipes(source_node.verbose)
        recv_pipes = self.get_recv_pipes(target_node.verbose)

        fail_count = 0
        count = 0
        target_datasets = []
        for source_dataset in source_datasets:

            # stats
            if self.args.progress:
                count = count + 1
                self.progress("Analysing dataset {}/{} ({} failed)".format(count, len(source_datasets), fail_count))

            try:
                # determine corresponding target_dataset
                target_name = self.make_target_name(source_dataset)
                target_dataset = target_node.get_dataset(target_name)
                target_datasets.append(target_dataset)

                # ensure parents exists
                # TODO: this isnt perfect yet, in some cases it can create parents when it shouldn't.
                if not self.args.no_send \
                        and target_dataset.parent \
                        and target_dataset.parent not in target_datasets \
                        and not target_dataset.parent.exists:
                    target_dataset.debug("Creating unmountable parents")
                    target_dataset.parent.create_filesystem(parents=True)

                # determine common zpool features (cached, so no problem we call it often)
                source_features = source_node.get_pool(source_dataset).features
                target_features = target_node.get_pool(target_dataset).features
                common_features = source_features and target_features

                # sync the snapshots of this dataset
                source_dataset.sync_snapshots(target_dataset, show_progress=self.args.progress,
                                              features=common_features, filter_properties=self.filter_properties_list(),
                                              set_properties=self.set_properties_list(),
                                              ignore_recv_exit_code=self.args.ignore_transfer_errors,
                                              holds=not self.args.no_holds, rollback=self.args.rollback,
                                              also_other_snapshots=self.args.other_snapshots,
                                              no_send=self.args.no_send,
                                              destroy_incompatible=self.args.destroy_incompatible,
                                              send_pipes=send_pipes, recv_pipes=recv_pipes,
                                              decrypt=self.args.decrypt, encrypt=self.args.encrypt,
                                              zfs_compressed=self.args.zfs_compressed, force=self.args.force, guid_check=not self.args.no_guid_check)
            except Exception as e:

                fail_count = fail_count + 1
                source_dataset.error("FAILED: " + str(e))
                if self.args.debug:
                    self.verbose("Debug mode, aborting on first error")
                    raise


        target_path_dataset = target_node.get_dataset(self.args.target_path)
        if not self.args.no_thinning:
            self.thin_missing_targets(target_dataset=target_path_dataset, used_target_datasets=target_datasets)

        if self.args.destroy_missing is not None:
            self.destroy_missing_targets(target_dataset=target_path_dataset, used_target_datasets=target_datasets)

        return fail_count

    def thin_source(self, source_datasets):

        self.set_title("Thinning source")

        for source_dataset in source_datasets:
            source_dataset.thin(skip_holds=True)

    def filter_properties_list(self):

        if self.args.filter_properties:
            filter_properties = self.args.filter_properties.split(",")
        else:
            filter_properties = []

        if self.args.clear_refreservation:
            filter_properties.append("refreservation")

        return filter_properties

    def set_properties_list(self):

        if self.args.set_properties:
            set_properties = self.args.set_properties.split(",")
        else:
            set_properties = []

        if self.args.clear_mountpoint:
            set_properties.append("canmount=noauto")

        return set_properties

    def set_snapshot_properties_list(self):

        if self.args.set_snapshot_properties:
            set_snapshot_properties = self.args.set_snapshot_properties.split(",")
        else:
            set_snapshot_properties = []

        return set_snapshot_properties

    def run(self):

        try:

            ################ create source zfsNode
            self.set_title("Source settings")

            description = "[Source]"
            if self.args.no_thinning:
                source_thinner = None
            else:
                source_thinner = Thinner(self.args.keep_source)
            source_node = ZfsNode(utc=self.args.utc,
                                  snapshot_time_format=self.snapshot_time_format, hold_name=self.hold_name, logger=self,
                                  ssh_config=self.args.ssh_config,
                                  ssh_to=self.args.ssh_source, readonly=self.args.test,
                                  debug_output=self.args.debug_output, description=description, thinner=source_thinner,
                                  exclude_snapshot_patterns=self.args.exclude_snapshot_pattern)

            ################# select source datasets
            self.set_title("Selecting")
            ( source_datasets, excluded_datasets) = source_node.selected_datasets(property_name=self.property_name,
                                                            exclude_received=self.args.exclude_received,
                                                            exclude_paths=self.exclude_paths,
                                                            exclude_unchanged=self.args.exclude_unchanged)
            if not source_datasets and not excluded_datasets:
                self.print_error_sources()
                return 255

            ################# snapshotting
            if not self.args.no_snapshot:
                self.set_title("Snapshotting")
                snapshot_name = datetime_now(self.args.utc).strftime(self.snapshot_time_format)
                source_node.consistent_snapshot(source_datasets, snapshot_name,
                                                min_changed_bytes=self.args.min_change,
                                                pre_snapshot_cmds=self.args.pre_snapshot_cmd,
                                                post_snapshot_cmds=self.args.post_snapshot_cmd,
                                                set_snapshot_properties=self.set_snapshot_properties_list())

            ################# sync
            # if target is specified, we sync the datasets, otherwise we just thin the source. (e.g. snapshot mode)
            if self.args.target_path:

                # create target_node
                self.set_title("Target settings")
                if self.args.no_thinning:
                    target_thinner = None
                else:
                    target_thinner = Thinner(self.args.keep_target)
                target_node = ZfsNode(utc=self.args.utc,
                                      snapshot_time_format=self.snapshot_time_format, hold_name=self.hold_name,
                                      logger=self, ssh_config=self.args.ssh_config,
                                      ssh_to=self.args.ssh_target,
                                      readonly=self.args.test, debug_output=self.args.debug_output,
                                      description="[Target]",
                                      thinner=target_thinner)
                target_node.verbose("Receive datasets under: {}".format(self.args.target_path))

                self.set_title("Synchronising")

                # check if exists, to prevent vague errors
                target_dataset = target_node.get_dataset(self.args.target_path)
                if not target_dataset.exists:
                    raise (Exception(
                        "Target path '{}' does not exist. Please create this dataset first.".format(target_dataset)))

                # check for collisions due to strip-path
                self.check_target_names(source_node, source_datasets, target_node)

                # do the actual sync
                # NOTE: even with no_send, no_thinning and no_snapshot it does a usefull thing because it checks if the common snapshots and shows incompatible snapshots
                fail_count = self.sync_datasets(
                    source_node=source_node,
                    source_datasets=source_datasets,
                    target_node=target_node)

            # no target specified, run in snapshot-only mode
            else:
                if not self.args.no_thinning:
                    self.thin_source(source_datasets)
                fail_count = 0

            if not fail_count:
                if self.args.test:
                    self.set_title("All tests successful.")
                else:
                    self.set_title("All operations completed successfully")
                    if not self.args.target_path:
                        self.verbose("(No target_path specified, only operated as snapshot tool.)")

            else:
                if fail_count != 255:
                    self.error("{} dataset(s) failed!".format(fail_count))

            if self.args.test:
                self.verbose("")
                self.warning("TEST MODE - DID NOT MAKE ANY CHANGES!")

            self.clear_progress()
            return fail_count

        except Exception as e:
            self.error("Exception: " + str(e))
            if self.args.debug:
                raise
            return 255
        except KeyboardInterrupt:
            self.error("Aborted")
            return 255


def cli():
    import sys

    signal(SIGPIPE, sigpipe_handler)

    failed_datasets=ZfsAutobackup(sys.argv[1:], False).run()
    sys.exit(min(failed_datasets, 255))


if __name__ == "__main__":
    cli()



import setuptools
from zfs_autobackup.ZfsAutobackup import ZfsAutobackup
import os

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="zfs_autobackup",
    version=ZfsAutobackup.VERSION,
    author="Edwin Eefting",
    author_email="edwin@datux.nl",
    description="ZFS autobackup is used to periodicly backup ZFS filesystems to other locations. It tries to be the most friendly to use and easy to debug ZFS backup tool.",
    long_description=long_description,
    long_description_content_type="text/markdown",

    url="https://github.com/psy0rz/zfs_autobackup",
    entry_points={
        'console_scripts':
            [
                'zfs-autobackup = zfs_autobackup.ZfsAutobackup:cli',
                'zfs-autoverify = zfs_autobackup.ZfsAutoverify:cli',
                'zfs-check = zfs_autobackup.ZfsCheck:cli',
            ]
    },
    packages=setuptools.find_packages(),

    classifiers=[
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=2.7',
    install_requires=[
        "colorama",
        "argparse"
    ]
)




from basetest import *


class TestZfsNode(unittest2.TestCase):

    def setUp(self):
        prepare_zpools()
        self.longMessage=True

    def test_keepsource0target10queuedsend(self):
        """Test if thinner doesnt destroy too much early on if there are no common snapshots YET. Issue #84"""

        with mocktime("20101111000000"):
            self.assertFalse(ZfsAutobackup(
                "test test_target1 --no-progress --verbose --keep-source=0 --keep-target=10 --allow-empty --no-send".split(
                    " ")).run())

        with mocktime("20101111000001"):
            self.assertFalse(ZfsAutobackup(
                "test test_target1 --no-progress --verbose --keep-source=0 --keep-target=10 --allow-empty --no-send".split(
                    " ")).run())

        with mocktime("20101111000002"):
            self.assertFalse(ZfsAutobackup(
                "test test_target1 --no-progress --verbose --keep-source=0 --keep-target=10 --allow-empty".split(
                    " ")).run())

        r = shelltest("zfs list -H -o name -r -t all " + TEST_POOLS)
        self.assertMultiLineEqual(r, """
test_source1
test_source1/fs1
test_source1/fs1@test-20101111000002
test_source1/fs1/sub
test_source1/fs1/sub@test-20101111000002
test_source2
test_source2/fs2
test_source2/fs2/sub
test_source2/fs2/sub@test-20101111000002
test_source2/fs3
test_source2/fs3/sub
test_target1
test_target1/test_source1
test_target1/test_source1/fs1
test_target1/test_source1/fs1@test-20101111000000
test_target1/test_source1/fs1@test-20101111000001
test_target1/test_source1/fs1@test-20101111000002
test_target1/test_source1/fs1/sub
test_target1/test_source1/fs1/sub@test-20101111000000
test_target1/test_source1/fs1/sub@test-20101111000001
test_target1/test_source1/fs1/sub@test-20101111000002
test_target1/test_source2
test_target1/test_source2/fs2
test_target1/test_source2/fs2/sub
test_target1/test_source2/fs2/sub@test-20101111000000
test_target1/test_source2/fs2/sub@test-20101111000001
test_target1/test_source2/fs2/sub@test-20101111000002
""")


    def test_excludepaths(self):
        """Test issue #103"""

        shelltest("zfs create test_target1/target_shouldnotbeexcluded")
        shelltest("zfs set autobackup:test=true test_target1/target_shouldnotbeexcluded")
        shelltest("zfs create test_target1/target")

        with mocktime("20101111000000"):
            self.assertFalse(ZfsAutobackup(
                "test test_target1/target --no-progress --verbose --allow-empty".split(
                    " ")).run())


        r = shelltest("zfs list -H -o name -r -t all " + TEST_POOLS)
        self.assertMultiLineEqual(r, """
test_source1
test_source1/fs1
test_source1/fs1@test-20101111000000
test_source1/fs1/sub
test_source1/fs1/sub@test-20101111000000
test_source2
test_source2/fs2
test_source2/fs2/sub
test_source2/fs2/sub@test-20101111000000
test_source2/fs3
test_source2/fs3/sub
test_target1
test_target1/target
test_target1/target/test_source1
test_target1/target/test_source1/fs1
test_target1/target/test_source1/fs1@test-20101111000000
test_target1/target/test_source1/fs1/sub
test_target1/target/test_source1/fs1/sub@test-20101111000000
test_target1/target/test_source2
test_target1/target/test_source2/fs2
test_target1/target/test_source2/fs2/sub
test_target1/target/test_source2/fs2/sub@test-20101111000000
test_target1/target/test_target1
test_target1/target/test_target1/target_shouldnotbeexcluded
test_target1/target/test_target1/target_shouldnotbeexcluded@test-20101111000000
test_target1/target_shouldnotbeexcluded
test_target1/target_shouldnotbeexcluded@test-20101111000000
""")





from basetest import *
from zfs_autobackup.CmdPipe import CmdPipe,CmdItem


class TestCmdPipe(unittest2.TestCase):

    def test_single(self):
        """single process stdout and stderr"""
        p=CmdPipe(readonly=False, inp=None)
        err=[]
        out=[]
        p.add(CmdItem(["sh", "-c", "echo out1;echo err1 >&2; echo out2; echo err2 >&2"], stderr_handler=lambda line: err.append(line), exit_handler=lambda exit_code: self.assertEqual(exit_code,0), stdout_handler=lambda line: out.append(line)))
        executed=p.execute()

        self.assertEqual(out, ["out1", "out2"])
        self.assertEqual(err, ["err1","err2"])
        self.assertIsNone(executed)

    def test_input(self):
        """test stdinput"""
        p=CmdPipe(readonly=False, inp="test")
        err=[]
        out=[]
        p.add(CmdItem(["cat"], stderr_handler=lambda line: err.append(line), exit_handler=lambda exit_code: self.assertEqual(exit_code,0), stdout_handler=lambda line: out.append(line) ))
        executed=p.execute()

        self.assertEqual(err, [])
        self.assertEqual(out, ["test"])
        self.assertIsNone(executed)

    def test_pipe(self):
        """test piped"""
        p=CmdPipe(readonly=False)
        err1=[]
        err2=[]
        err3=[]
        out=[]
        p.add(CmdItem(["echo", "test"], stderr_handler=lambda line: err1.append(line), exit_handler=lambda exit_code: self.assertEqual(exit_code,0)))
        p.add(CmdItem(["tr", "e", "E"], stderr_handler=lambda line: err2.append(line), exit_handler=lambda exit_code: self.assertEqual(exit_code,0)))
        p.add(CmdItem(["tr", "t", "T"], stderr_handler=lambda line: err3.append(line), exit_handler=lambda exit_code: self.assertEqual(exit_code,0), stdout_handler=lambda line: out.append(line)))
        executed=p.execute()

        self.assertEqual(err1, [])
        self.assertEqual(err2, [])
        self.assertEqual(err3, [])
        self.assertEqual(out, ["TEsT"])
        self.assertIsNone(executed)

        #test str representation as well
        self.assertEqual(str(p), "(echo test) | (tr e E) | (tr t T)")

    def test_pipeerrors(self):
        """test piped stderrs """
        p=CmdPipe(readonly=False)
        err1=[]
        err2=[]
        err3=[]
        out=[]
        p.add(CmdItem(["sh", "-c", "echo err1 >&2"], stderr_handler=lambda line: err1.append(line), ))
        p.add(CmdItem(["sh", "-c", "echo err2 >&2"], stderr_handler=lambda line: err2.append(line), ))
        p.add(CmdItem(["sh", "-c", "echo err3 >&2"], stderr_handler=lambda line: err3.append(line), stdout_handler=lambda line: out.append(line)))
        executed=p.execute()

        self.assertEqual(err1, ["err1"])
        self.assertEqual(err2, ["err2"])
        self.assertEqual(err3, ["err3"])
        self.assertEqual(out, [])
        self.assertTrue(executed)

    def test_exitcode(self):
        """test piped exitcodes """
        p=CmdPipe(readonly=False)
        err1=[]
        err2=[]
        err3=[]
        out=[]
        p.add(CmdItem(["sh", "-c", "exit 1"], stderr_handler=lambda line: err1.append(line), exit_handler=lambda exit_code: self.assertEqual(exit_code,1)))
        p.add(CmdItem(["sh", "-c", "exit 2"], stderr_handler=lambda line: err2.append(line), exit_handler=lambda exit_code: self.assertEqual(exit_code,2)))
        p.add(CmdItem(["sh", "-c", "exit 3"], stderr_handler=lambda line: err3.append(line), exit_handler=lambda exit_code: self.assertEqual(exit_code,3), stdout_handler=lambda line: out.append(line)))
        executed=p.execute()

        self.assertEqual(err1, [])
        self.assertEqual(err2, [])
        self.assertEqual(err3, [])
        self.assertEqual(out, [])
        self.assertIsNone(executed)

    def test_readonly_execute(self):
        """everything readonly, just should execute"""

        p=CmdPipe(readonly=True)
        err1=[]
        err2=[]
        out=[]

        def true_exit(exit_code):
            return True

        p.add(CmdItem(["echo", "test1"], stderr_handler=lambda line: err1.append(line), exit_handler=true_exit, readonly=True))
        p.add(CmdItem(["echo", "test2"], stderr_handler=lambda line: err2.append(line), exit_handler=true_exit, readonly=True, stdout_handler=lambda line: out.append(line)))
        executed=p.execute()

        self.assertEqual(err1, [])
        self.assertEqual(err2, [])
        self.assertEqual(out, ["test2"])
        self.assertTrue(executed)

    def test_readonly_skip(self):
        """one command not readonly, skip"""

        p=CmdPipe(readonly=True)
        err1=[]
        err2=[]
        out=[]
        p.add(CmdItem(["echo", "test1"], stderr_handler=lambda line: err1.append(line), readonly=False))
        p.add(CmdItem(["echo", "test2"], stderr_handler=lambda line: err2.append(line), readonly=True, stdout_handler=lambda line: out.append(line)))
        executed=p.execute()

        self.assertEqual(err1, [])
        self.assertEqual(err2, [])
        self.assertEqual(out, [])
        self.assertTrue(executed)

    def test_no_handlers(self):
        with self.assertRaises(Exception):
            p=CmdPipe()
            p.add(CmdItem([ "echo" ]))
            p.execute()

        #NOTE: this will give some resource warnings

    def test_manual_pipes(self):

        # manual piping means: a command in the pipe has a stdout_handler, which is responsible for sending the data into the next item of the pipe.

        result=[]


        def stdout_handler(line):
            item2.process.stdin.write(line.encode('utf8'))

            # item2.process.stdin.close()

        item1=CmdItem(["echo", "test"], stdout_handler=stdout_handler)
        item2=CmdItem(["tr", "e", "E"], stdout_handler=lambda line: result.append(line))

        p=CmdPipe()
        p.add(item1)
        p.add(item2)
        p.execute()

        self.assertEqual(result, ["tEst"])

    def test_multiprocess(self):

        #dont do any piping at all, just run multiple processes and handle outputs

        result1=[]
        result2=[]
        result3=[]

        item1=CmdItem(["echo", "test1"], stdout_handler=lambda line: result1.append(line))
        item2=CmdItem(["echo", "test2"], stdout_handler=lambda line: result2.append(line))
        item3=CmdItem(["echo", "test3"], stdout_handler=lambda line: result3.append(line))

        p=CmdPipe()
        p.add(item1)
        p.add(item2)
        p.add(item3)
        p.execute()

        self.assertEqual(result1, ["test1"])
        self.assertEqual(result2, ["test2"])
        self.assertEqual(result3, ["test3"])




from __future__ import print_function

import time
from signal import signal, SIGPIPE

from . import util
from .TreeHasher import TreeHasher
from .BlockHasher import BlockHasher
from .ZfsNode import ZfsNode
from .util import *
from .CliBase import CliBase


class ZfsCheck(CliBase):

    def __init__(self, argv, print_arguments=True):

        # NOTE: common options argument parsing are in CliBase
        super(ZfsCheck, self).__init__(argv, print_arguments)

        self.node = ZfsNode(self.log, utc=self.args.utc, readonly=self.args.test, debug_output=self.args.debug_output)

        self.block_hasher = BlockHasher(count=self.args.count, bs=self.args.block_size, skip=self.args.skip)

    def get_parser(self):

        parser = super(ZfsCheck, self).get_parser()

        # positional arguments
        parser.add_argument('target', metavar='TARGET', default=None, nargs='?', help='Target to checkum. (can be blockdevice, directory or ZFS snapshot)')

        group = parser.add_argument_group('Checker options')

        group.add_argument('--block-size', metavar="BYTES", default=4096, help="Read block-size, default %(default)s",
                           type=int)
        group.add_argument('--count', metavar="COUNT", default=int((100 * (1024 ** 2)) / 4096),
                           help="Hash chunks of COUNT blocks. Default %(default)s . (CHUNK size is BYTES * COUNT) ", type=int)  # 100MiB

        group.add_argument('--check', '-c', metavar="FILE", default=None, const=True, nargs='?',
                           help="Read hashes from STDIN (or FILE) and compare them")

        group.add_argument('--skip', '-s', metavar="NUMBER", default=0, type=int,
                           help="Skip this number of chunks after every hash. %(default)s")

        return parser

    def parse_args(self, argv):
        args = super(ZfsCheck, self).parse_args(argv)

        if args.test:
            self.warning("TEST MODE - WILL ONLY DO READ-ONLY STUFF")

        if args.target is None:
            self.error("Please specify TARGET")
            sys.exit(1)

        self.verbose("Target               : {}".format(args.target))
        self.verbose("Block size           : {} bytes".format(args.block_size))
        self.verbose("Block count          : {}".format(args.count))
        self.verbose("Effective chunk size : {} bytes".format(args.count*args.block_size))
        self.verbose("Skip chunk count     : {} (checks {:.2f}% of data)".format(args.skip, 100/(1+args.skip)))
        self.verbose("")


        return args

    def prepare_zfs_filesystem(self, snapshot):

        mnt = "/tmp/" + tmp_name()
        self.debug("Create temporary mount point {}".format(mnt))
        self.node.run(["mkdir", mnt])
        snapshot.mount(mnt)
        return mnt

    def cleanup_zfs_filesystem(self, snapshot):
        mnt = "/tmp/" + tmp_name()
        snapshot.unmount(mnt)
        self.debug("Cleaning up temporary mount point")
        self.node.run(["rmdir", mnt], hide_errors=True, valid_exitcodes=[])

    # NOTE: https://www.google.com/search?q=Mount+Path+Limit+freebsd
    # Freebsd has limitations regarding path length, so we have to clone it so the part stays sort
    def prepare_zfs_volume(self, snapshot):
        """clone volume, waits and tries to findout /dev path to the volume, in a compatible way. (linux/freebsd/smartos)"""

        clone_name = get_tmp_clone_name(snapshot)
        clone = snapshot.clone(clone_name)

        # TODO: add smartos location to this list as well
        locations = [
            "/dev/zvol/" + clone_name
        ]

        clone.debug("Waiting for /dev entry to appear in: {}".format(locations))
        time.sleep(0.1)

        start_time = time.time()
        while time.time() - start_time < 10:
            for location in locations:
                if os.path.exists(location):
                    return location

                # fake it in testmode
                if self.args.test:
                    return location

            time.sleep(1)

        raise (Exception("Timeout while waiting for /dev entry to appear. (looking in: {}). Hint: did you forget to load the encryption key?".format(locations)))

    def cleanup_zfs_volume(self, snapshot):
        """destroys temporary volume snapshot"""
        clone_name = get_tmp_clone_name(snapshot)
        clone = snapshot.zfs_node.get_dataset(clone_name)
        clone.destroy(deferred=True, verbose=False)

    def generate_tree_hashes(self, prepared_target):

        tree_hasher = TreeHasher(self.block_hasher)
        self.debug("Hashing tree: {}".format(prepared_target))
        for i in tree_hasher.generate(prepared_target):
            yield i

    def generate_tree_compare(self, prepared_target, input_generator=None):

        tree_hasher = TreeHasher(self.block_hasher)
        self.debug("Comparing tree: {}".format(prepared_target))
        for i in tree_hasher.compare(prepared_target, input_generator):
            yield i

    def generate_file_hashes(self, prepared_target):

        self.debug("Hashing file: {}".format(prepared_target))
        for i in self.block_hasher.generate(prepared_target):
            yield i

    def generate_file_compare(self, prepared_target, input_generator=None):

        self.debug("Comparing file: {}".format(prepared_target))
        for i in self.block_hasher.compare(prepared_target, input_generator):
            yield i

    def generate_input(self):
        """parse input lines and yield items to use in compare functions"""

        if self.args.check is True:
            input_fh=sys.stdin
        else:
            input_fh=open(self.args.check, 'r')

        last_progress_time = time.time()
        progress_checked = 0
        progress_skipped = 0

        line=input_fh.readline()
        skip=0
        while line:
            i=line.rstrip().split("\t")
            #ignores lines without tabs
            if (len(i)>1):

                if skip==0:
                    progress_checked=progress_checked+1
                    yield i
                    skip=self.args.skip
                else:
                    skip=skip-1
                    progress_skipped=progress_skipped+1

                if self.args.progress and time.time() - last_progress_time > 1:
                    last_progress_time = time.time()
                    self.progress("Checked {} hashes (skipped {})".format(progress_checked, progress_skipped))

            line=input_fh.readline()

        self.verbose("Checked {} hashes (skipped {})".format(progress_checked, progress_skipped))

    def print_hashes(self, hash_generator):
        """prints hashes that are yielded by the specified hash_generator"""

        last_progress_time = time.time()
        progress_count = 0

        for i in hash_generator:

            if len(i) == 3:
                print("{}\t{}\t{}".format(*i))
            else:
                print("{}\t{}".format(*i))
            progress_count = progress_count + 1

            if self.args.progress and time.time() - last_progress_time > 1:
                last_progress_time = time.time()
                self.progress("Generated {} hashes.".format(progress_count))

            sys.stdout.flush()

        self.verbose("Generated {} hashes.".format(progress_count))
        self.clear_progress()

        return 0

    def print_errors(self, compare_generator):
        """prints errors that are yielded by the specified compare_generator"""
        errors = 0
        for i in compare_generator:
            errors = errors + 1

            if len(i) == 4:
                (file_name, chunk_nr, compare_hexdigest, actual_hexdigest) = i
                print("{}: Chunk {} failed: {} {}".format(file_name, chunk_nr, compare_hexdigest, actual_hexdigest))
            else:
                (chunk_nr, compare_hexdigest, actual_hexdigest) = i
                print("Chunk {} failed: {} {}".format(chunk_nr, compare_hexdigest, actual_hexdigest))

            sys.stdout.flush()

        self.verbose("Total errors: {}".format(errors))
        self.clear_progress()

        return errors

    def prepare_target(self):

        if "@" in self.args.target:
            # zfs snapshot
            snapshot=self.node.get_dataset(self.args.target)
            if not snapshot.exists:
                raise Exception("ZFS snapshot {} does not exist!".format(snapshot))
            dataset_type = snapshot.parent.properties['type']

            if dataset_type == 'volume':
                return self.prepare_zfs_volume(snapshot)
            elif dataset_type == 'filesystem':
                return self.prepare_zfs_filesystem(snapshot)
            else:
                raise Exception("Unknown dataset type")
        return self.args.target

    def cleanup_target(self):
        if "@" in self.args.target:
            # zfs snapshot
            snapshot=self.node.get_dataset(self.args.target)
            if not snapshot.exists:
                return

            dataset_type = snapshot.parent.properties['type']

            if dataset_type == 'volume':
                self.cleanup_zfs_volume(snapshot)
            elif dataset_type == 'filesystem':
                self.cleanup_zfs_filesystem(snapshot)

    def run(self):

        compare_generator=None
        hash_generator=None
        try:
            prepared_target=self.prepare_target()
            is_dir=os.path.isdir(prepared_target)

            #run as compare
            if self.args.check is not None:
                input_generator=self.generate_input()
                if is_dir:
                    compare_generator = self.generate_tree_compare(prepared_target, input_generator)
                else:
                    compare_generator=self.generate_file_compare(prepared_target, input_generator)
                errors=self.print_errors(compare_generator)
            #run as generator
            else:
                if is_dir:
                    hash_generator = self.generate_tree_hashes(prepared_target)
                else:
                    hash_generator=self.generate_file_hashes(prepared_target)

                errors=self.print_hashes(hash_generator)

        except Exception as e:
            self.error("Exception: " + str(e))
            if self.args.debug:
                raise
            return 255
        except KeyboardInterrupt:
            self.error("Aborted")
            return 255

        finally:
            #important to call check_output so that cleanup still functions in case of a broken pipe:
            # util.check_output()

            #close generators, to make sure files are not in use anymore when cleaning up
            if hash_generator is not None:
                hash_generator.close()
            if compare_generator is not None:
                compare_generator.close()
            self.cleanup_target()

        return errors


def cli():
    import sys
    signal(SIGPIPE, sigpipe_handler)
    failed=ZfsCheck(sys.argv[1:], False).run()
    sys.exit(min(failed,255))


if __name__ == "__main__":
    cli()



import hashlib
import os


class BlockHasher():
    """This class was created to checksum huge files and blockdevices (TB's)
    Instead of one sha1sum of the whole file, it generates sha1susms of chunks of the file.

    The chunksize is count*bs (bs is the read blocksize from disk)

    Its also possible to only read a certain percentage of blocks to just check a sample.

    Input and output generators are in the format ( chunk_nr, hexdigest )

    NOTE: skipping is only used on the generator side. The compare side just compares what it gets from the input generator.

    """

    def __init__(self, count=10000, bs=4096, hash_class=hashlib.sha1, skip=0):
        self.count = count
        self.bs = bs
        self.chunk_size=bs*count
        self.hash_class = hash_class

        # self.coverage=coverage
        self.skip=skip
        self._skip_count=0

        self.stats_total_bytes=0


    def _seek_next_chunk(self, fh, fsize):
        """seek fh to next chunk and update skip counter.
        returns chunk_nr
        return false it should skip the rest of the file


        """

        #ignore rempty files
        if fsize==0:
            return False

        # need to skip chunks?
        if self._skip_count > 0:
            chunks_left = ((fsize - fh.tell()) // self.chunk_size) + 1
            # not enough chunks left in this file?
            if self._skip_count >= chunks_left:
                # skip rest of this file
                self._skip_count = self._skip_count - chunks_left
                return False
            else:
                # seek to next chunk, reset skip count
                fh.seek(self.chunk_size * self._skip_count, os.SEEK_CUR)
                self._skip_count = self.skip
                return  fh.tell()//self.chunk_size
        else:
            # should read this chunk, reset skip count
            self._skip_count = self.skip
            return fh.tell() // self.chunk_size

    def generate(self, fname):
        """Generates checksums

        yields(chunk_nr, hexdigest)

        yields nothing for empty files.
        """


        with open(fname, "rb") as fh:

            fh.seek(0, os.SEEK_END)
            fsize=fh.tell()
            fh.seek(0)

            while fh.tell()<fsize:
                chunk_nr=self._seek_next_chunk(fh, fsize)
                if chunk_nr is False:
                    return

                #read chunk
                hash = self.hash_class()
                block_nr = 0
                while block_nr != self.count:
                    block=fh.read(self.bs)
                    if block==b"":
                        break
                    hash.update(block)
                    block_nr = block_nr + 1

                yield (chunk_nr, hash.hexdigest())

    def compare(self, fname, generator):
        """reads from generator and compares blocks
        Yields mismatches in the form: ( chunk_nr, hexdigest, actual_hexdigest)
        Yields errors in the form: ( chunk_nr, hexdigest, "message" )

        """

        try:
            checked = 0
            with open(fname, "rb") as f:
                for (chunk_nr, hexdigest) in generator:
                    try:

                        checked = checked + 1
                        hash = self.hash_class()
                        f.seek(int(chunk_nr) * self.bs * self.count)
                        block_nr = 0
                        for block in iter(lambda: f.read(self.bs), b""):
                            hash.update(block)
                            block_nr = block_nr + 1
                            if block_nr == self.count:
                                break

                        if block_nr == 0:
                            yield (chunk_nr, hexdigest, 'EOF')

                        elif (hash.hexdigest() != hexdigest):
                            yield (chunk_nr, hexdigest, hash.hexdigest())

                    except Exception as e:
                        yield ( chunk_nr , hexdigest, 'ERROR: '+str(e))

        except Exception as e:
            yield ( '-', '-', 'ERROR: '+ str(e))